
/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database 2)
// Generic performance evaluation model of OODBs
// Version 2.0 C++, April 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// basec.h: Database definition (Classes)
/////////////////////////////////////////////////////////////////////

class Database;
class Class;
class Object;
class Page;

/////////////////////////////////////////////////////////////////////
// CLASS Database
/////////////////////////////////////////////////////////////////////

class Database {

  public:                       // Methods

    Database();                 // Constructor
    ~Database();                // Destructor
    void Generation();          // Database generation
    void Display();             // Database display
    void Initp();               // Initial placement of objects
    void SeqPl();               // Sequential initial placement
    void OSeqPl();              // Optimized sequential initial placement
    void OtherPl();             // Other initial placement
    Class *GetClass(int i);     // Return class #i
    Object *GetObj(int i);      // Return object #i
    Page *NewPage();            // Create new page
    void AddPage(Page *page);   // Add page to list
    Page *GetPage(int num);     // Return pointer to page #num
    void DelPage(int num);      // Delete a page
    void PurgePages();          // Delete all pages
    void DisplayPages();        // Display used pages 
    int NbPages();              // Return the number of used pages

    // Attributes used by the DSTC clustering strategy
    // Made public to easen usage

    ChList *mo[MNO][2];                 // Obseravtion matrix MO
                                        // 0 - "j" index
                                        // 1 - counter value
    int vo[MNO];                        // Observation vector VO
    ChList *mc[MNO][4];                 // Consolidated matrix MC
                                        // 0 - "j" index
                                        // 1 - fc(i,j) factor value
                                        // 2 - update date
                                        // 3 - clustering flag
    int pi;                             // Current period
    int ni;                             // Number of current entries in MO

    // DSTC methods

    void NewPeriod();                   // New observation period
    void DisplayMO();                   // Display the MO matrix
    void DisplayMC();                   // Display the MC matrix

  private:                      // Attributes

    int nbclasses;              // Number of classes
    Class *classtable[MNC];     // Index of classes
    int nbobj;                  // Number of objects
    Object *objtable[MNO];      // Index of instances
    int nbpages;                // Number of pages
    Page *pagelist;             // Pointer toward the page chained list
    int cumulp;                 // Number of created pages (=> unique ids)
    short Inheritance[MNREFT];  // Inheritance hierarchy flag
    short NoCycle[MNREFT];      // Cyclic hierarchy flag

};

/////////////////////////////////////////////////////////////////////
// CLASS Class
/////////////////////////////////////////////////////////////////////

class Class {

  public:                       // Methods

    Class(int i, short maxn, int bs); // Constructor
    ~Class();                   // Destructor
    short MaxNRef();            // Return maxnref
    int BaseSize();             // Return basesize
    void SetTRef(int i, short t); // Initialize tref[i] to t
    short GetTRef(int i);       // Return tref[i]
    void SetCRef(int i, Class *c); // Initialize cref[i] to c
    Class *GetCRef(int i);      // Return cref[i]
    ChList *Iterator();         // Return the iterator's address
    void SetSize(int s);        // Initialize instsize to s
    int GetSize();              // Return intsize
    int Id();                   // Return id
    void Display();             // Screen display

  private:                      // Attributes

    int id;                     // Class identifier
    short maxnref;              // Maximum number of references
    int basesize;               // Instance base size
    short tref[MMAXNREF];       // Reference types table
    Class *cref[MMAXNREF];      // Referenced classes table
    ChList *iterator;           // Iterator (list of instance OIDs)
    int instsize;               // Instance size

};


/////////////////////////////////////////////////////////////////////
// CLASS Object
/////////////////////////////////////////////////////////////////////

class Object {

  public:                       // Methods

    Object(int id, Class *c);   // Constructr
    ~Object();                  // Destructr
    Class *ClassPtr();          // Return the object's class
    void SetORef(int i, Object *o); // Initialize oref[i] to o
    Object *GetORef(int i);     // Return oref[i]
    ChList *BackRef();          // Return the backward references address
    ChList *PageDisk();         // Return the list of pages containing the object
    int OID();                  // Return OID
    void Display();             // Screen display

    private:                    // Attributes

    int oid;                    // OID
    Class *classptr;            // Class
    Object *oref[MMAXNREF];     // Table of object references
    ChList *backref;            // List of backward references (OIDs)
    ChList *pagedisk;           // List of the pages containing the object

};

/////////////////////////////////////////////////////////////////////
// CLASS Page
/////////////////////////////////////////////////////////////////////

class Page {

  public:                       // Methods

    Page(int num);              // Constructor
    ~Page();                    // Destructor
    void SetNext(Page *n);      // Initialization of next pointer
    int PageNum();              // Return page number
    int Free();                 // Return free space
    void ResetFree(int f);      // Reinitialize free space
    ChList *Objects();          // Return list of objects
    Page *Next();               // Return next page

  private:                      // Attributes

    int pagenum;                // Page number
//  int sector;                 // Sector number
//  int cyl;                    // Cylinder number
    int free;                   // Free space in page
    ChList *objects;            // Objects stored in page
    Page *next;                 // Next page in the list

};

