
/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database 2)
// Generic performance evaluation model of OODBs
// Version 2.0 C++, April 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// basem.h: Database definition (Methods)
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// CLASS Database
/////////////////////////////////////////////////////////////////////

// CLASS Database: Constructor

Database::Database() {

  nbclasses=0;
  nbobj=0;
  nbpages=0;
  cumulp=0;
  pagelist=NULL;

  // DSTC

  int i, j;

  pi=0;                                 // 1st period
  ni=0;
  // Initialization of MO & MC
  for (i=0; i<NO; i++) {
    for (j=0; j<2; j++)
      mo[i][j]=new ChList;
    vo[i]=0;
    for (j=0; j<4; j++)
      mc[i][j]=new ChList;
  }
}

// CLASS Database: Destructor

Database::~Database() {

  int i, j;

  for (i=0; i<nbclasses; i++) delete classtable[i];
  for (i=0; i<nbobj; i++) delete objtable[i];
  PurgePages();
  for (i=0; i<NO; i++) {
    for (j=0; j<2; j++)
      delete mo[i][j];
    for (j=0; j<4; j++)
      delete mc[i][j];
  }
}

// CLASS Database: Database generation 

void Database::Generation() {

  int i, j, k, l, count;
  short pb;
  ChList *pile;
  Class *cclass;

//  printf("\nDatabase generation...\n");

  // Preliminary step:
  // initialization of cyclic and inheritance hierarchy flags

  NoCycle[0]=1;                         // RefType #0 => inheritance
  Inheritance[0]=1;
  for (i=1; i<NREFT; i++) {
    NoCycle[i]=0;
//    NoCycle[i]=VALUE TO SET-UP;
    Inheritance[i]=0;
//    Inheritance[i]=VALUE TO SET-UP;
  }

  // Step #1: Schema generation - classes

  nbclasses=0;
  for (i=0; i<NC; i++) {
    // Instantiation of class Class
    classtable[i]=new Class(i,MAXNREF,BASESIZE);
    nbclasses++;
    // Determination of refetence type for each reference
    for (j=0; j<classtable[i]->MaxNRef(); j++)
      classtable[i]->SetTRef(j,CRand(DIST1,0,NREFT-1));
  }

  // Step #2: Inter-class links

  for (i=0; i<NC; i++)
    for (j=0; j<classtable[i]->MaxNRef(); j++)
      // Reference affectation
      classtable[i]->SetCRef(j,classtable[CRand(DIST2,INFCLASS-1,SUPCLASS-1)]);
//      classtable[i]->SetCRef(j,classtable[VALUE TO SET-UP]);

  // Step #3: Acyclic graphs consistency

  for (i=0; i<NC; i++)
    for (j=0; j<classtable[i]->MaxNRef(); j++)
      if ((NoCycle[classtable[i]->GetTRef(j)])
      && (classtable[i]->GetCRef(j)!=NULL)) {
        // Traversal of CRef(j) class hierarchy
        pile=new ChList;
        pile->Insert(1,classtable[i]->GetCRef(j)->Id());
        count=1;                        // Cycle counter
        pb=0;                           // No problemo
        while ((pile->NbElem()) && (!pb)) {
          cclass=classtable[pile->Scan(1)];
          pile->Delete(1);
          count++;
          if ((cclass==classtable[i]) || (count>=NC))
            // Class in hierarchy or cycle detected
            pb=1;
//          else
            // Corresponding non-null references are stacked
            for (k=0; k<cclass->MaxNRef(); k++)
              if ((cclass->GetTRef(k)==classtable[i]->GetTRef(j))
              && (cclass->GetCRef(k)!=NULL))
                pile->Insert(1,cclass->GetCRef(k)->Id());
          }
        delete pile;
        if (pb) classtable[i]->SetCRef(j,NULL);
        else if (Inheritance[classtable[i]->GetTRef(j)]) { // Inheritance
          // Update instance size in the inhertance graph
          pile=new ChList;
          pile->Insert(1,classtable[i]->GetCRef(j)->Id());
          while (pile->NbElem()) {
            cclass=classtable[pile->Scan(1)];
            pile->Delete(1);
            cclass->SetSize(cclass->GetSize()+classtable[i]->BaseSize());
            for (k=0; k<cclass->MaxNRef(); k++)
              if ((cclass->GetTRef(k)==classtable[i]->GetTRef(j))
              && (cclass->GetCRef(k)!=NULL))
                pile->Insert(1,cclass->GetCRef(k)->Id());
          }
          delete pile;
          }
        }

  // Step #4: Instances

  nbobj=0;
  for (i=0; i<NO; i++) {
    // Instantiation of class Object
    objtable[i]=new Object(i,classtable[CRand(DIST3,0,NC-1)]);
    nbobj++;
    // Addition into class iterator
    objtable[i]->ClassPtr()->Iterator()->Insert(1,i);
  }


  // Step #5: Inter-instance links

  for (i=0; i<NO; i++)
    for (j=0; j<objtable[i]->ClassPtr()->MaxNRef(); j++) {
      // Object reference to NULL
      objtable[i]->SetORef(j,NULL);
      // If class referenced by mother class not null
      if (objtable[i]->ClassPtr()->GetCRef(j)!=NULL)
        // If iterator not empty
        if (objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->NbElem()) {
          // Reference affectation 
          l=CRand(DIST4,1,objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->NbElem());
          objtable[i]->SetORef(j,objtable[objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->Scan(l)]);
          // Backward reference affectation for the referenced object
          objtable[objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->Scan(l)]->BackRef()->Insert(1,objtable[i]->OID());
        }
    }
}

// CLASS Database: Database display

void Database::Display() {

  int i, j;
  float sizec, refc, ref[MNREFT], refn[MNREFT];

  printf("\n*** Database Schema ***\n");
  sizec=0;
  refc=0;
  for (i=0; i<NREFT; i++) {
    ref[i]=0;
    refn[i]=0;
  }
  for (i=0; i<NC; i++) {
    classtable[i]->Display();
    sizec+=classtable[i]->GetSize();
    for (j=0; j<classtable[i]->MaxNRef(); j++) {
      ref[classtable[i]->GetTRef(j)]++;
      if (classtable[i]->GetCRef(j)!=NULL) {
        refc++;
        refn[classtable[i]->GetTRef(j)]++;
      }
    }
  }
  printf("\n Instances mean size                 : %f\n",sizec/NC);
  printf(" Mean number of references per class : %f\n",refc/NC);
  for (i=0; i<NREFT; i++) {
  printf(" Mean number of type #%d references   : %f (%f are not NULL)\n",i,ref[i]/NC,refn[i]/NC);
  }

  printf("\n*** Database Instantiation ***\n");
  refc=0;
  for (i=0; i<NO; i++) {
    objtable[i]->Display();
    for (j=0; j<objtable[i]->ClassPtr()->MaxNRef(); j++)
      if (objtable[i]->GetORef(j)!=NULL) refc++;
  }
  printf("\n Mean number of references per object: %f\n",refc/NO);
}

// CLASS Database: Initial placement of objects

void Database::Initp() {

//  printf("Objects initial placement [%d]... ",INITP);

  switch(INITP) {
    case 1: SeqPl();break;      // Sequential initial placement
    case 2: OSeqPl();break;     // Optimized sequential initial placement
    case 9: OtherPl();break;    // Other initial placement
  }

//  printf("%d pages used\n",nbpages);
}

// CLASS Database: Sequential initial placement

void Database::SeqPl() {

  Page *cur;
  int i, rsize;

  // Creation of a new page and insertion into the list
  cur=NewPage();
  AddPage(cur);
  // Placement of each object
  for (i=0; i<NO; i++) {
    if (objtable[i]->ClassPtr()->GetSize()<=PGSIZE) {
      // Normal object, size <= page size
      if (objtable[i]->ClassPtr()->GetSize()>cur->Free()) {
        // Size > free space => new page
        cur=NewPage();
        AddPage(cur);
      }
      // Object inserted in the page content
      cur->Objects()->Insert(1,i);
      // Page into object list of pages
      objtable[i]->PageDisk()->Insert(1,cur->PageNum());
      // Free space update
      cur->ResetFree(cur->Free()-objtable[i]->ClassPtr()->GetSize());
    } else { // Large object
      // rsize: size remaining
      rsize=objtable[i]->ClassPtr()->GetSize()-cur->Free();
      // Current page full
      cur->ResetFree(0);
      // Object inserted in the page content
      cur->Objects()->Insert(1,i);
      // Page into object list of pages
      objtable[i]->PageDisk()->Insert(1,cur->PageNum());
      // While the object is not wholly stored
      while (rsize>0) {
        // New page
        cur=NewPage();
        AddPage(cur);
        // Object inserted in the page content
        cur->Objects()->Insert(1,i);
        // Page into object list of pages
        objtable[i]->PageDisk()->Insert(1,cur->PageNum());
        if (cur->Free()<rsize) {   // Object does not fit
          rsize=rsize-cur->Free(); // Remaining size --
          cur->ResetFree(0);       // Page full
        } else {
          cur->ResetFree(cur->Free()-rsize); // Free space --
          rsize=0;
        }
      }
    }
  }
}

// CLASS Database: Optimized sequential initial placement

void Database::OSeqPl() {

  Page *cur, *goodpage, *scan;
  int i, rsize;
  short space;

  // Creation of a new page and insertion into the list
  cur=NewPage();
  AddPage(cur);
  // Placement of each object
  for (i=0; i<NO; i++) {
    if (objtable[i]->ClassPtr()->GetSize()<=PGSIZE) {
      // Normal object, size <= page size
      if (objtable[i]->ClassPtr()->GetSize()<=cur->Free())
        // Size <= free space
        goodpage=cur;
      else { // Look for a candidate page among existing pages
        scan=pagelist;
        space=0;
        while ((!space) && (scan!=NULL))
          if (objtable[i]->ClassPtr()->GetSize()<=scan->Free())
            space=1;
          else
            scan=scan->Next();
        if (space) goodpage=scan;
        else { // New page
          cur=NewPage();
          AddPage(cur);
          goodpage=cur;
        }
      }
      // Object inserted in the page content
      goodpage->Objects()->Insert(1,i);
      // Page into object list of pages
      objtable[i]->PageDisk()->Insert(1,goodpage->PageNum());
      // Free space update
      goodpage->ResetFree(goodpage->Free()-objtable[i]->ClassPtr()->GetSize());
    } else { // Large object
      // rsize: remaining size
      rsize=objtable[i]->ClassPtr()->GetSize()-cur->Free();
      // Current page full
      cur->ResetFree(0);
      // Object inserted in the page content
      cur->Objects()->Insert(1,i);
      // Page into object list of pages
      objtable[i]->PageDisk()->Insert(1,cur->PageNum());
      // While object not fully stored
      while (rsize>0) {
        // New page
        cur=NewPage();
        AddPage(cur);
        // Object inserted in the page content
        cur->Objects()->Insert(1,i);
        // Page into object list of pages
        objtable[i]->PageDisk()->Insert(1,cur->PageNum());
        if (cur->Free()<rsize) {   // Object does not fit
          rsize=rsize-cur->Free(); // Remaining size --
          cur->ResetFree(0);       // Page full
        } else {
          cur->ResetFree(cur->Free()-rsize); // Free space --
          rsize=0;
        }
      }
    }
  }
}

// CLASS Database: Other initial placement (user-defined)

void Database::OtherPl() {

  // INSERT CODE HERE
}

// CLASS Database: Return class #i

Class *Database::GetClass(int i) {

  if ((i>=0) && (i<nbclasses)) return classtable[i];
  else return NULL;
}

// CLASS Database: Return object #i

Object *Database::GetObj(int i) {

  if ((i>=0) && (i<nbobj)) return objtable[i];
  else return NULL;
}

// CLASS Database: Create new page

Page *Database::NewPage() {

  Page *newpage;
  newpage=new Page(cumulp++);
  return newpage;
}

// CLASS Database: Add page into list (head)

void Database::AddPage(Page *page) {

  page->SetNext(pagelist);
  pagelist=page;
  nbpages++;
}

// CLASS Database: Return ptr toward page #num

Page *Database::GetPage(int num) {

  Page *cur;

  cur=pagelist;
  while ((cur!=NULL) && (cur->PageNum()!=num)) cur=cur->Next();
  if (cur!=NULL) return cur;
  else return NULL;
}

// CLASS Database: Delete page #num

void Database::DelPage(int num) {

  Page *cur, *prev;

  prev=NULL;
  cur=pagelist;
  while ((cur!=NULL) && (cur->PageNum()!=num)) {
    prev=cur;
    cur=cur->Next();
  }
  if ((cur!=NULL) && (cur->PageNum()==num)) { // Suppression
    if (prev!=NULL) prev->SetNext(cur->Next());
    else pagelist=cur->Next();
    delete cur;
  }
  nbpages--;
}

// CLASS Database: Delete all pages

void Database::PurgePages() {

  Page *cur, *del;

  cur=pagelist;
  del=NULL;
  while (cur!=NULL) {
    del=cur;
    cur=cur->Next();
    delete del;
  }
  pagelist=NULL;
}

// CLASS Database: Display used pages 

void Database::DisplayPages() {

  Page *cur;
  int i;

  printf("\n*** Disk pages used ***\n");
  cur=pagelist;
  while (cur!=NULL) {
    printf("\nPage #%d\n",cur->PageNum());
    printf("\tFree space: %d\n",cur->Free());
    printf("\tObjects contained:");
    for (i=1; i<=cur->Objects()->NbElem(); i++)
      printf(" %d",cur->Objects()->Scan(i));
    printf("\n");
    cur=cur->Next();
  }
}

// CLASS Database: Return the number of used pages 

int Database::NbPages() {

  return nbpages;
}


// CLASS Database: New observation period

void Database::NewPeriod() {

  int i, j;

  // Next period
  pi++;
  if (pi>=P) pi=0;
  // Reset MO
  ni=0;
  for (i=0; i<NO; i++) {
    for (j=0; j<2; j++) {
      delete mo[i][j];
      mo[i][j]=new ChList;
    }
    vo[i]=0;
  }
}

// CLASS Database: Display matrix MO

void Database::DisplayMO() {

  int i, j;

  printf("\n*** MO ***\n\n");
  for (i=0; i<NO; i++) {
    printf("%d)",i);
    if (mo[i][0]->NbElem()==mo[i][1]->NbElem())
      for (j=1; j<=mo[i][0]->NbElem(); j++)
        printf(" %d:%d",mo[i][0]->Scan(j),mo[i][1]->Scan(j));
    else                              // Probleme !!
      printf(" ERROR: Discrepancy in MO!");
    printf(" - %d\n",vo[i]);
  }
}

// CLASS Database: Display matrix MC

void Database::DisplayMC() {

  int i, j;

  printf("\n*** MC ***\n\n");
  for (i=0; i<NO; i++) {
    printf("%d)",i);
    for (j=1; j<=mc[i][0]->NbElem(); j++)
      printf(" %d:%d/%d/%d",mc[i][0]->Scan(j),mc[i][1]->Scan(j),
      mc[i][2]->Scan(j),mc[i][3]->Scan(j));
    printf("\n");
  }
}

/////////////////////////////////////////////////////////////////////
// CLASS Class
/////////////////////////////////////////////////////////////////////

// CLASS Class: Constructor

Class::Class(int i, short maxn, int bs) {

  id=i;
  maxnref=maxn;
  basesize=bs;
  iterator=new ChList;
  instsize=basesize;  // O2: basesize + X (~700)
}

// CLASS Class: Destructor

Class::~Class() {

  delete iterator;
}

// CLASS Class: Return maxnref

short Class::MaxNRef() {

  return maxnref;
}

// CLASS Class: Return basesize

int Class::BaseSize() {

  return basesize;
}

// CLASS Class: Initialize tref[i] to t

void Class::SetTRef(int i, short t) {

  if ((i>=0) && (i<maxnref)) tref[i]=t;
}

// CLASS Class: Return tref[i]

short Class::GetTRef(int i) {

  if ((i>=0) && (i<maxnref)) return tref[i];
  else return -1;
}

// CLASS Class: Initialize cref[i] to c

void Class::SetCRef(int i, Class *c) {

  if ((i>=0) && (i<maxnref)) cref[i]=c;
}

// CLASS Class: Return cref[i]

Class *Class::GetCRef(int i) {

  if ((i>=0) && (i<maxnref)) return cref[i];
  else return NULL;
}

// CLASS Class: Return iterator address

ChList *Class::Iterator() {

  return iterator;
}

// CLASS Class: Initialize instsize to s

void Class::SetSize(int s) {

  instsize=s;
}

// CLASS Class: Return intsize

int Class::GetSize() {

  return instsize;
}

// CLASS Class: Return id

int Class::Id() {

  return id;
}

// CLASS Class: Screen display

void Class::Display() {

  int i;

  printf("\n*** Class #%d\n",id);
  printf("\tmaxnref=%d\n",maxnref);
  printf("\tbasesize=%d\n",basesize);
  printf("\tcref(tref)=");
  for (i=0; i<maxnref; i++)
    if (cref[i]!=NULL) printf("%d(%d) ",cref[i]->Id(),tref[i]);
    else printf("NULL(%d) ",tref[i]);
  printf("\n\titerator=");
  for (i=1; i<=iterator->NbElem(); i++) printf("%d ",iterator->Scan(i));
  printf("\n\tinstsize=%d\n",instsize);
}

/////////////////////////////////////////////////////////////////////
// CLASS Object
/////////////////////////////////////////////////////////////////////

// CLASS Object: Constructor

Object::Object(int id, Class *c) {

  oid=id;
  classptr=c;
  backref=new ChList;
  pagedisk=new ChList;
}

// CLASS Object: Destructor

Object::~Object() {

  delete backref;
  delete pagedisk;
}

// CLASS Object: Return object class

Class *Object::ClassPtr() {

  return classptr;
}

// CLASS Object: Initialize oref[i] to o

void Object::SetORef(int i, Object *o) {

  if ((i>=0) && (i<classptr->MaxNRef())) oref[i]=o;
}

// CLASS Object: Return oref[i]

Object *Object::GetORef(int i) {

  if ((i>=0) && (i<classptr->MaxNRef())) return oref[i];
  else return NULL;
}

// CLASS Object: Return backward references address

ChList *Object::BackRef() {

  return backref;
}

// CLASS Object: Return the list of pages containing the object

ChList *Object::PageDisk() {

  return pagedisk;
}

// CLASS Object: Return OID

int Object::OID() {

  return oid;
}

// CLASS Object: Screen display

void Object::Display() {

  int i;

  printf("\n*** Object #%d\n",oid);
  printf("\tclass=%d\n",classptr->Id());
  printf("\toref=");
  for (i=0; i<classptr->MaxNRef(); i++)
    if (oref[i]!=NULL) printf("%d ",oref[i]->OID());
    else printf("NULL ");
  printf("\n\tbackref=");
  for (i=1; i<=backref->NbElem(); i++) printf("%d ",backref->Scan(i));
  printf("\n\tpagedisk=");
  for (i=1; i<=pagedisk->NbElem(); i++) printf("%d ",pagedisk->Scan(i));
  printf("\n");
}

/////////////////////////////////////////////////////////////////////
// CLASS Page
/////////////////////////////////////////////////////////////////////

// CLASS Page: Constructor

Page::Page(int num) {

  pagenum=num;
  free=PGSIZE;
  objects=new ChList;
}

// CLASS Page: Destructor

Page::~Page() {

  delete objects;
}

// CLASS Page: Next pointer initialization 

void Page::SetNext(Page *n) {

  next=n;
}

// CLASS Page: Return page number

int Page::PageNum() {

  return pagenum;
}

// CLASS Page: Return free space

int Page::Free() {

  return free;
}
// CLASS Page: Reinitialize free space

void Page::ResetFree(int f) {

  free=f;
}

// CLASS Page: Return list of stored objects

ChList *Page::Objects() {

  return objects;
}

// CLASS Page: Return next page 

Page *Page::Next() {

  return next;
}

