
/////////////////////////////////////////////////////////////////////
// DESP-C++ (Discrete-Event #Simulation Package)
// Version 1.1, February 1998
// Jerome Darmont
// LIMOS, Blaise Pascal University (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// eventc.h: Event #manager methods deENDition 
// Varies with the simulated system
// VOODB2 set-up (April 1998)
/////////////////////////////////////////////////////////////////////

#include "stdlib.h"
#include "string.h"

/////////////////////////////////////////////////////////////////////
// CLASS EventManager
/////////////////////////////////////////////////////////////////////

// CLASS EventManager: Constructor

EventManager::EventManager(Simulation *sim) {

  int i;
  char name[STRS], num[STRS];

  simul=sim;

  // Instantiation of passive resources

  // Server processor
  serverpr=new Resource("Server Processor",1,simul);

  // Client processors
  if (SYSCLASS==1) // Centralized architecture 
    // No clients, all the treatments are performed on the same machine
    for (i=0; i<NUSERS; i++) clientpr[i]=serverpr;
  else if ((SYSCLASS>=2) && (SYSCLASS<=4))
    for (i=0; i<NUSERS; i++) {
      strcpy(name,"Client #");
      // itoa(i,num,10);
      sprintf(num,"%d",i);
      strcat(name,num);
      strcat(name," Processor");
      clientpr[i]=new Resource(name,1,simul);
    }
  else if (SYSCLASS==9) {  // User-deENDed architecture 
    // Insert code here
  }

  // Server disk
  disk=new Resource("Disk",1,simul);

  // Database
  db=new Resource("Database",MULTILVL,simul);

  // Instantiation of active resources

  // Clients
  for (i=0; i<NUSERS; i++) {
    strcpy(name,"User #");
    //itoa(i,num,10);
    sprintf(num,"%d",i);
    strcat(name,num);
    user[i]=new User(name,1,simul,i,clientpr[i]);
  }

  // Transaction manager
  if (SYSCLASS==4)              // DB server
    transmgr=new TransMgr("Transaction Mgr",INENDITE,simul,serverpr,db,1/NETTHRU);
  else if (SYSCLASS<=3)
    transmgr=new TransMgr("Transaction Mgr",INENDITE,simul,clientpr[0],db,0);
  else {                        // User-deENDed architecture 
    // Insert code here
  }

  // Clustering manager
  if (SYSCLASS==4)              // DB server
    clustmgr=new ClustMgr("Clustering Mgr",INENDITE,simul,serverpr,db);
  else if (SYSCLASS<=3)
    clustmgr=new ClustMgr("Clustering Mgr",INENDITE,simul,clientpr[0],db);
  else {                        // User-deENDed architecture 
    // Insert code here
  }

  // Object manager
  if ((SYSCLASS==1) || (SYSCLASS==4))   // Centralized system or DB server
    objmgr=new ObjMgr("Object Mgr",INENDITE,simul,serverpr,0);
  else if (SYSCLASS==2)                 // Object server
    objmgr=new ObjMgr("Object Mgr",INENDITE,simul,serverpr,1/NETTHRU);
  else if (SYSCLASS==3)                 // Page server
    objmgr=new ObjMgr("Object Mgr",INENDITE,simul,clientpr[0],0);
  else {                                // User-deENDed architecture 
    // Insert code here
  }
  clustmgr->SetObjPtr(objmgr);

  // Buffering manager
  if (SYSCLASS==3)                      // Page server
    buffmgr=new BuffMgr("Buffering Mgr",INENDITE,simul,serverpr,PGSIZE/NETTHRU);
  else if ((SYSCLASS<=2) || (SYSCLASS==4))
    buffmgr=new BuffMgr("Buffering Mgr",INENDITE,simul,serverpr,0);
  else {                                // User-deENDed architecture 
    // Insert code here
  }
  clustmgr->SetBuffPtr(buffmgr);

  // I/O subsystem
  if (SYSCLASS!=9) iosubsys=new IOSubSys("I/O Subsystem",INENDITE,simul,serverpr,disk);
  else {                                // User-deENDed architecture 
    // Insert code here
  }
}

// CLASS EventManager: Destructor

EventManager::~EventManager() {

  int i;

  // Destruction of passive resources 
  delete serverpr;
  for (i=0; i<NUSERS; i++) delete clientpr[i];
  delete disk;
  delete db;

  // Destruction of active resources 
  for (i=0; i<NUSERS; i++) delete user[i];
  delete transmgr;
  delete clustmgr;
  delete objmgr;
  delete buffmgr;
  delete iosubsys;
}

// CLASS EventManager: Event #execution 

void EventManager::ExecuteEvent(int code, Client *client) {

  int i;
  Client *newcli;

//  printf("Performing Event ##%d at time %f\n",code,simul->Tnow());

  switch(code) {

  // Initial event: creation of one client per user

  case 0:{ client->cliid=0;
           client->type=0;
           user[client->cliid]->Event10(client);
           for (i=1; i<NUSERS; i++) {
             newcli=simul->NewClient();
             newcli->cliid=i;
             client->type=0;
             user[newcli->cliid]->Event10(newcli);
           }
           break;
         }

  // Current events

  case 10: user[client->cliid]->Event10(client);break;
  case 11: user[client->cliid]->Event11(client);break;
  case 12: user[client->cliid]->Event12(client);break;
  case 13: user[client->cliid]->Event13(client);break;

  case 20: transmgr->Event20(client);break;
  case 21: transmgr->Event21(client);break;
  case 22: transmgr->Event22(client);break;
  case 23: transmgr->Event23(client);break;
  case 24: transmgr->Event24(client);break;
  case 25: transmgr->Event25(client);break;
  case 26: transmgr->Event26(client);break;
  case 27: transmgr->Event27(client);break;
  case 28: transmgr->Event28(client);break;

  case 30: clustmgr->Event30(client);break;
  case 31: clustmgr->Event31(client);break;
  case 32: clustmgr->Event32(client);break;
  case 33: clustmgr->Event33(client);break;
  case 34: clustmgr->Event34(client);break;
  case 35: clustmgr->Event35(client);break;
  case 36: clustmgr->Event36(client);break;
  case 37: clustmgr->Event37(client);break;
  case 38: clustmgr->Event38(client);break;

  case 40: objmgr->Event40(client);break;
  case 41: objmgr->Event41(client);break;
  case 42: objmgr->Event42(client);break;
  case 43: objmgr->Event43(client);break;

  case 50: buffmgr->Event50(client);break;
  case 51: buffmgr->Event51(client);break;
  case 52: buffmgr->Event52(client);break;
  case 53: buffmgr->Event53(client);break;
  case 54: buffmgr->Event54(client);break;
  case 55: buffmgr->Event55(client);break;

  case 60: iosubsys->Event60(client);break;
  case 61: iosubsys->Event61(client);break;
  case 62: iosubsys->Event62(client);break;
  case 63: iosubsys->Event63(client);break;

  default: printf("Error: unknown Event ##%d at time %f\n",code,simul->Tnow());
  }
}

// CLASS EventManager: Initialization of stats for each resource

void EventManager::Init() {

  int i;

  // Passive resources
  serverpr->ResetStats();
  for (i=0; i<NUSERS; i++) clientpr[i]->ResetStats();
  disk->ResetStats();
  db->ResetStats();

  // Active resources
  for (i=0; i<NUSERS; i++) user[i]->ResetStats();
  transmgr->ResetStats();
  clustmgr->ResetStats();
  objmgr->ResetStats();
  buffmgr->ResetStats();
  iosubsys->ResetStats();

  // VOODB2-specific stats
  for (i=0; i<NTTRANS+1; i++) {
    NBTRANSC[i]=0;
    NBTRANSC2[i]=0;
    RTIMEC[i]=0;
    RTIMEC2[i]=0;
    NBOBJC[i]=0;
    NBOBJC2[i]=0;
  }
  NBCLC=0;
  NBCLC2=0;
  CTIMEC=0;
  CTIMEC2=0;
  TCTIMEC=0;
  TCTIMEC2=0;
  NBIOC=0;
  NBIOC2=0;
  TNBIOC=0;
  TNBIOC2=0;
  CNBIOC=0;
  CNBIOC2=0;
  NBPGC=0;
  NBPGC2=0;
  NCC=0;
  NCC2=0;
  TCC=0;
  TCC2=0;
}

// CLASS EventManager: Replications initialization 

void EventManager::InitRep() {

  int i;

  // Database generation & objects initial placement

  DB=new Database;
  DB->Generation();
  DB->Initp();
//  DB->Display();
//  DB->DisplayPages();

  // Scheduler
  simul->Sched()->Purge();

  // Passive resources
  serverpr->ResetCounters();
  serverpr->PurgeQueue();
  for (i=0; i<NUSERS; i++) {
    clientpr[i]->ResetCounters();
    clientpr[i]->PurgeQueue();
  }
  disk->ResetCounters();
  disk->PurgeQueue();
  db->ResetCounters();
  db->PurgeQueue();

  // Active resources
  for (i=0; i<NUSERS; i++) {
    user[i]->ResetCounters();
    user[i]->PurgeQueue();
  }
  transmgr->ResetCounters();
  transmgr->PurgeQueue();
  clustmgr->ResetCounters();
  clustmgr->PurgeQueue();
  objmgr->ResetCounters();
  objmgr->PurgeQueue();
  buffmgr->ResetCounters();
  buffmgr->PurgeQueue();
  buffmgr->InitBuffer();
  iosubsys->ResetCounters();
  iosubsys->PurgeQueue();

  // VOODB2-specific stats
  for (i=0; i<NTTRANS+1; i++) {
    NBTRANS[i]=0;
    RTIME[i]=0;
    NBST[i]=0;
    NBOBJ[i]=0;
  }
  NBCL=0;
  CTIME=0;
  TCTIME=0;
  NBIO=0;
  TNBIO=0;
  CNBIO=0;
  NCLUSTERS=0;
  TCLUSTERS=0;
}

// CLASS EventManager: Stat computation for each resource

void EventManager::Stats() {

  int i;
  float rtimeinc;

  // Passive resources
  serverpr->Stats();
  for (i=0; i<NUSERS; i++) clientpr[i]->Stats();
  disk->Stats();
  db->Stats();

  // Active resources
  for (i=0; i<NUSERS; i++) user[i]->Stats();
  transmgr->Stats();
  clustmgr->Stats();
  objmgr->Stats();
  buffmgr->Stats();
  iosubsys->Stats();

  // VOODB2-specific stats
  for (i=0; i<NTTRANS+1; i++) {
    NBTRANSC[i]+=NBTRANS[i];
    NBTRANSC2[i]+=NBTRANS[i]*NBTRANS[i];
    if (NBTRANS[i]!=0) rtimeinc=(RTIME[i]+(NBST[i]-NBTRANS[i])*simul->Tnow())/NBTRANS[i];
    else rtimeinc=0;
    RTIMEC[i]+=rtimeinc;
    RTIMEC2[i]+=rtimeinc*rtimeinc;
    if (NBTRANS[i]!=0) {
      NBOBJC[i]+=NBOBJ[i]/NBTRANS[i];
      NBOBJC2[i]+=(NBOBJ[i]/NBTRANS[i])*(NBOBJ[i]/NBTRANS[i]);
    } else {
      NBOBJC[i]=0;
      NBOBJC2[i]=0;
    }
  }
  NBCLC+=NBCL;
  NBCLC2+=NBCL*NBCL;
  CTIMEC+=CTIME;
  CTIMEC2+=CTIME*CTIME;
  TCTIMEC+=TCTIME;
  TCTIMEC2+=TCTIME*TCTIME;
  NBIOC+=NBIO;
  NBIOC2+=NBIO*NBIO;
  TNBIOC+=TNBIO;
  TNBIOC2+=TNBIO*TNBIO;
  CNBIOC+=CNBIO;
  CNBIOC2+=CNBIO*CNBIO;
  NBPGC+=DB->NbPages();
  NBPGC2+=DB->NbPages()*DB->NbPages();
  NCC+=NCLUSTERS;
  NCC2+=NCLUSTERS*NCLUSTERS;
  if (NCLUSTERS>0) {
    TCC+=TCLUSTERS/NCLUSTERS;
    TCC2+=(TCLUSTERS*TCLUSTERS)/(NCLUSTERS*NCLUSTERS);
  }
  
  // Destruction of database at the end of the replication
  delete DB;
}

// CLASS EventManager: Display stats for each resource

void EventManager::DisplayStats() {

  int i;

  printf("\n\n*** SIMULATION STATISTICS ***\n");

  //

  printf("\n*** PASSIVE RESOURCES\n");
  serverpr->DisplayStats();
  for (i=0; i<NUSERS; i++) clientpr[i]->DisplayStats();
  disk->DisplayStats();
  db->DisplayStats();

  //

  printf("\n*** ACTIVE RESOURCES\n");
  for (i=0; i<NUSERS; i++) user[i]->DisplayStats();
  transmgr->DisplayStats();
  clustmgr->DisplayStats();
  objmgr->DisplayStats();
  buffmgr->DisplayStats();
  iosubsys->DisplayStats();

  // VOODB2-specific stats

  printf("\n*** TRANSACTIONS\n\n");

  //

  printf("Mean number of transactions\n\n");

  float nbtransm[NTTRANS+1], nbtransd[NTTRANS+1], nbtransi[NTTRANS+1];

  for (i=0; i<NTTRANS+1; i++) {
    nbtransm[i]=NBTRANSC[i]/NREPLIC;
    nbtransd[i]=(NREPLIC*NBTRANSC2[i]-NBTRANSC[i]*NBTRANSC[i])/(NREPLIC*NREPLIC);
    if (nbtransd[i]>0) nbtransd[i]=sqrt(nbtransd[i]);
    else nbtransd[i]=0;
    if (NREPLIC>1) nbtransi[i]=t(NREPLIC-1)*nbtransd[i]/sqrt(NREPLIC);
    else nbtransi[i]=0;
  }

  printf("\t* Set-oriented accesses : %10.2f\t+/- %10.2f\n",nbtransm[0],nbtransi[0]);
  printf("\t* Simple traversals     : %10.2f\t+/- %10.2f\n",nbtransm[1],nbtransi[1]);
  printf("\t* Hierarchy traversals  : %10.2f\t+/- %10.2f\n",nbtransm[2],nbtransi[2]);
  printf("\t* Stochastic traversals : %10.2f\t+/- %10.2f\n",nbtransm[3],nbtransi[3]);
  printf("\t* TOTAL                 : %10.2f\t+/- %10.2f\n",nbtransm[4],nbtransi[4]);

  //

  printf("\nMean response time\n\n");

  float mrt[NTTRANS+1], drt[NTTRANS+1], irt[NTTRANS+1];

  for (i=0; i<NTTRANS; i++) {
    mrt[i]=RTIMEC[i]/NREPLIC;
    drt[i]=(NREPLIC*RTIMEC2[i]-RTIMEC[i]*RTIMEC[i])/(NREPLIC*NREPLIC);
    if (drt[i]>0) drt[i]=sqrt(drt[i]);
    else drt[i]=0;
    if (NREPLIC>1) irt[i]=t(NREPLIC-1)*drt[i]/sqrt(NREPLIC);
    else irt[i]=0;    
  }
  if (nbtransm[NTTRANS]!=0) mrt[NTTRANS]=(transmgr->Mean(0)*transmgr->Mean(2))/nbtransm[NTTRANS];
  else mrt[NTTRANS]=0;
  if (nbtransd[NTTRANS]!=0) drt[NTTRANS]=(transmgr->Dev(0)*transmgr->Dev(2))/nbtransd[NTTRANS];
  else drt[NTTRANS]=0;
  if (NREPLIC>1) irt[NTTRANS]=t(NREPLIC-1)*drt[NTTRANS]/sqrt(NREPLIC);
  else irt[NTTRANS]=0;    
    
  printf("\t* Set-oriented accesses : %10.2f ms\t+/- %10.2f\n",mrt[0],irt[0]);
  printf("\t* Simple traversals     : %10.2f ms\t+/- %10.2f\n",mrt[1],irt[1]);
  printf("\t* Hierarchy traversals  : %10.2f ms\t+/- %10.2f\n",mrt[2],irt[2]);
  printf("\t* Stochastic traversals : %10.2f ms\t+/- %10.2f\n",mrt[3],irt[3]);
  printf("\t* TOTAL                 : %10.2f ms\t+/- %10.2f\n",mrt[4],irt[4]);

  //

  printf("\nMean number of objects accessed\n\n");

  float nbobjm[NTTRANS+1], nbobjd[NTTRANS+1], nbobji[NTTRANS+1];

  for (i=0; i<NTTRANS+1; i++) {
    nbobjm[i]=NBOBJC[i]/NREPLIC;
    nbobjd[i]=(NREPLIC*NBOBJC2[i]-NBOBJC[i]*NBOBJC[i])/(NREPLIC*NREPLIC);
    if (nbobjd[i]>0) nbobjd[i]=sqrt(nbobjd[i]);
    else nbobjd[i]=0;
    if (NREPLIC>1) nbobji[i]=t(NREPLIC-1)*nbobjd[i]/sqrt(NREPLIC);
    else nbobji[i]=0;      
  }

  printf("\t* Set-oriented accesses : %10.2f\t+/- %10.2f\n",nbobjm[0],nbobji[0]);
  printf("\t* Simple traversals     : %10.2f\t+/- %10.2f\n",nbobjm[1],nbobji[1]);
  printf("\t* Hierarchy traversals  : %10.2f\t+/- %10.2f\n",nbobjm[2],nbobji[2]);
  printf("\t* Stochastic traversals : %10.2f\t+/- %10.2f\n",nbobjm[3],nbobji[3]);
  printf("\t* TOTAL                 : %10.2f\t+/- %10.2f\n",nbobjm[4],nbobji[4]);

  //

  printf("\n*** CLUSTERING\n\n");

  float nbclm=NBCLC/NREPLIC;
  float nbcld=(NREPLIC*NBCLC2-NBCLC*NBCLC)/(NREPLIC*NREPLIC);
  if (nbcld>0) nbcld=sqrt(nbcld);
  else nbcld=0;
  float nbcli;
  if (NREPLIC>1) nbcli=t(NREPLIC-1)*nbcld/sqrt(NREPLIC);
  else nbcli=0;   
  //
  float ctimem=0;
  if (nbclm!=0) ctimem=CTIMEC/(nbclm*NREPLIC*1000);
  float ctimed=(NREPLIC*CTIMEC2-CTIMEC*CTIMEC)/(NREPLIC*NREPLIC*1000);
  if (ctimed>0) ctimed=sqrt(ctimed);
  else ctimed=0;
  if (nbcld!=0) ctimed=ctimed/nbcld;
  float ctimei;
  if (NREPLIC>1) ctimei=t(NREPLIC-1)*ctimed/sqrt(NREPLIC);
  else ctimei=0;  
  // ctimed false measure!
  //
  float tctimem=TCTIMEC/(NREPLIC*1000);
  float tctimed=(NREPLIC*TCTIMEC2-TCTIMEC*TCTIMEC)/(NREPLIC*NREPLIC*1000);
  if (tctimed>0) tctimed=sqrt(tctimed);
  else tctimed=0;
  float tctimei;
  if (NREPLIC>1) tctimei=t(NREPLIC-1)*tctimed/sqrt(NREPLIC);
  else tctimei=0;  
  //
  float nccm=NCC/NREPLIC;
  float nccd=(NREPLIC*NCC2-NCC*NCC)/(NREPLIC*NREPLIC);
  if (nccd>0) nccd=sqrt(nccd); else nccd=0;
  float ncci;
  if (NREPLIC>1) ncci=t(NREPLIC-1)*nccd/sqrt(NREPLIC);
  else ncci=0;  
  //
  float tccm=TCC/NREPLIC;
  float tccd=(NREPLIC*TCC2-TCC*TCC)/(NREPLIC*NREPLIC);
  if (tccd>0) tccd=sqrt(tccd); else tccd=0;
  float tcci;
  if (NREPLIC>1) tcci=t(NREPLIC-1)*tccd/sqrt(NREPLIC);
  else tcci=0;  

  printf("\t* Mean number of reclusterings  : %10.2f  \t+/- %10.2f\n",nbclm,nbcli);
  printf("\t* Mean clustering duration      : %10.2f s\t+/- %10.2f\n",ctimem,ctimei);
  printf("\t* Mean cl.-rel. ops. overhead   : %10.2f s\t+/- %10.2f\n",tctimem,tctimei);
  printf("\t* Mean number of clusters built : %10.2f \t+/- %10.2f\n",nccm,ncci);
  printf("\t* Clusters' average size        : %10.2f o\t+/- %10.2f\n",tccm,tcci);

  //

  printf("\n*** THROUHPUT\n\n");

  float thrum=nbtransm[4]*1000/simul->Tmax();
//  float thrud=nbtransd[4]*1000/simul->Tmax(); // ?
  float thrui=nbtransi[4]*1000/simul->Tmax(); // ?  

  printf("\t* Mean Throughput : %10.2f tr./s\t+/- %10.2f\n",thrum,thrui);

  //

  printf("\n*** I/Os\n\n");

  float nbiom=NBIOC/NREPLIC;
  float nbiod=(NREPLIC*NBIOC2-NBIOC*NBIOC)/(NREPLIC*NREPLIC);
  if (nbiod>0) nbiod=sqrt(nbiod);
  else nbiod=0;
  float nbioi;
  if (NREPLIC>1) nbioi=t(NREPLIC-1)*nbiod/sqrt(NREPLIC);
  else nbioi=0;    
  //
  float tnbiom=TNBIOC/NREPLIC;
  float tnbiod=(NREPLIC*TNBIOC2-TNBIOC*TNBIOC)/(NREPLIC*NREPLIC);
  if (tnbiod>0) tnbiod=sqrt(tnbiod);
  else tnbiod=0;
  float tnbioi;
  if (NREPLIC>1) tnbioi=t(NREPLIC-1)*tnbiod/sqrt(NREPLIC);
  else tnbioi=0;   
  //
  float cnbiom=CNBIOC/NREPLIC;
  float cnbiod=(NREPLIC*CNBIOC2-CNBIOC*CNBIOC)/(NREPLIC*NREPLIC);
  if (cnbiod>0) cnbiod=sqrt(cnbiod);
  else cnbiod=0;
  float cnbioi;
  if (NREPLIC>1) cnbioi=t(NREPLIC-1)*cnbiod/sqrt(NREPLIC);
  else cnbioi=0;   

  printf("\t* Mean number of I/Os : %10.2f\t+/- %10.2f\n",nbiom,nbioi);
  printf("\t* Transaction I/Os    : %10.2f\t+/- %10.2f\n",tnbiom,tnbioi);
  printf("\t* Clustering I/Os     : %10.2f\t+/- %10.2f\n",cnbiom,cnbioi);

  //

  printf("\n*** NUMBER OF PAGES\n\n");

  float nbpgm=NBPGC/NREPLIC;
  float nbpgd=(NREPLIC*NBPGC2-NBPGC*NBPGC)/(NREPLIC*NREPLIC);
  if (nbpgd>0) nbpgd=sqrt(nbpgd);
  else nbpgd=0;
  float nbpgi;
  if (NREPLIC>1) nbpgi=t(NREPLIC-1)*nbpgd/sqrt(NREPLIC);
  else nbpgi=0; 
  
  printf("\t* Mean number of pages used : %10.2f\t+/- %10.2f\n\n",nbpgm,nbpgi);
}

/////////////////////////////////////////////////////////////////////
// CLASS User (AR1)
/////////////////////////////////////////////////////////////////////

// CLASS User: Constructor

User::User(char n[STRS], int cap, Simulation *sim, short id, Resource *rsp): Resource(n, cap, sim) {

  userid=id;
  rsproc=rsp;
}

// CLASS User: Event ##0 - Arrival on AR1

void User::Event10(Client *client) {

  this->P(11,client,1);         // Reservation of AR1
}

// CLASS User: Event ##1 - Latency time

void User::Event11(Client *client) {

  Sim()->Sched()->Schedule(12,Sim()->Tnow()+Exp(THINK),client);
}

// CLASS User: Event ##2 - Wait for processor

void User::Event12(Client *client) {

  rsproc->P(13,client,1);
}

// CLASS User: Event ##3 - Transaction generation 

void User::Event13(Client *client) {

  Client *newcli;

  newcli=GenTrans();            // R1
  rsproc->V();                  // Release processor
  if (newcli->type<5)           // Transaction
    Sim()->Sched()->Schedule(20,Sim()->Tnow(),newcli);
  else                          // Clustering
    Sim()->Sched()->Schedule(30,Sim()->Tnow(),newcli);
  Sim()->Sched()->Schedule(10,Sim()->Tnow(),client); // New transaction
  this->V();                    // Release AR1
}

// CLASS User: R1 - Generate transaction

Client *User::GenTrans() {

  Client *newcli;
  float p;

  // New client
  newcli=Sim()->NewClient();
  newcli->cliid=userid;

  // Root object 
  newcli->rootoid=CRand(DIST5,0,NO-1);
  newcli->trans_it=0;

  // Previous object (DSTC)
  newcli->prevoid=-1;

  // Transaction type selection
  p=Uni(0,1);
  if (p<=PCT1) {                // Set-oriented access
    newcli->type=1;
    newcli->curdepth=SETDEPTH;
  } else if (p<=PCT2) {         // Simple traversal
    newcli->type=2;
    newcli->curdepth=SIMDEPTH;
  } else if (p<=PCT3) {         // Hierarchy traversal
    newcli->type=3;
    newcli->reftype=CRand(1,0,NREFT-1);
    newcli->curdepth=HIEDEPTH;
  } else if (p<=PCC) {          // Stochastic traversal
    newcli->type=4;
    newcli->curdepth=STODEPTH;
  } else newcli->type=5;        // Clustering demand

//  printf("Transaction (type #%d) created at time %f\n",newcli->type,Sim()->Tnow());

  return newcli;
}

/////////////////////////////////////////////////////////////////////
// CLASS TransMgr (AR2)
/////////////////////////////////////////////////////////////////////

// CLASS TransMgr: Constructor

TransMgr::TransMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsb, float t): Resource(n, cap, sim) {

  rsproc[0]=rsp;
  rsbd=rsb;
  tnet=t;
}

// CLASS TransMgr: Event #0 - Arrival on AR2

void TransMgr::Event20(Client *client) {

  this->P(21,client,1);         // Reservation AR2
}

// CLASS TransMgr: Event #1 - Wait for processor (1)

void TransMgr::Event21(Client *client) {

  if (client->trans_it==0) {
    NBST[client->type-1]++;
    RTIME[client->type-1]-=Sim()->Tnow();
  }
  client->trans_it++;
  rsproc[client->cliid]->P(22,client,1);
}

// CLASS TransMgr: Event #2 - Extract object / reserve DB

void TransMgr::Event22(Client *client) {

  ExtObj(client);               // R2
  rsproc[client->cliid]->V();   // Release processor
  if (client->curdepth>=0)      // Transaction in process
    rsbd->P(23,client,1);       // Reserve DB
  else {
    // Count completed transactions
    NBTRANS[client->type-1]++;
    NBTRANS[NTTRANS]++;
    RTIME[client->type-1]+=Sim()->Tnow();
    this->V();                  // Release AR2
    Sim()->KillClient(client);  // Destroy current client
  }
}

// CLASS TransMgr: Event #3 - Wait for processor (2)

void TransMgr::Event23(Client *client) {

  rsproc[client->cliid]->P(24,client,1);
}

// CLASS TransMgr: Event #4 - Acquire locks

void TransMgr::Event24(Client *client) {

  Sim()->Sched()->Schedule(25,Sim()->Tnow()+GETLOCK,client);
}

// CLASS TransMgr: Event #5 - Object request

void TransMgr::Event25(Client *client) {

  rsproc[client->cliid]->V();   // Release processor
  // Object request
  Sim()->Sched()->Schedule(40,Sim()->Tnow(),client); // Request transmitted
}

// CLASS TransMgr: Event #6 - Wait for processor (3)

void TransMgr::Event26(Client *client) {

  rsproc[client->cliid]->P(27,client,1);
}

// CLASS TransMgr: Event #7 - Treatment of transaction & locks release

void TransMgr::Event27(Client *client) {

  TrTrans();                    // R6
  Sim()->Sched()->Schedule(28,Sim()->Tnow()+RELLOCK,client); // Release locks
}

// CLASS TransMgr: Event #8 - Network transfer

void TransMgr::Event28(Client *client) {

  rsproc[client->cliid]->V();   // Release processor
  rsbd->V();                    // Release DB
  this->V();                    // Release AR2
  Sim()->Sched()->Schedule(30,Sim()->Tnow()+tnet*DB->GetObj(client->rootoid)->ClassPtr()->GetSize(),client); // Network transfer
}

// CLASS TransMgr: R2 - Extract object

void TransMgr::ExtObj(Client *client) {

  switch(client->type) {

    case 1: NextSet(client);break;      // Set-oriented access
    case 2: NextSimp(client);break;     // Simple traversal
    case 3: NextHier(client);break;     // Hierarchy traversal
    case 4: NextStoc(client);break;     // Stochastic traversal

    default: client->curdepth=-1;       // Error => STOP
  }

//  printf("Transaction type #%d - Depth %d, Next object: #%d (at time %f)\n",client->type,client->curdepth,client->rootoid,Sim()->Tnow());
}

// CLASS TransMgr: R2.1 - Set-oriented access

void TransMgr::NextSet(Client *client) {

  int i, j, k, n;

  if (client->curdepth==SETDEPTH) {     // 1st passage
    // Creation of list roots
    client->roots=new ChList();
    // Creation of list toaccess with rootoid
    client->toaccess=new ChList();
    client->toaccess->Insert(1,client->rootoid);
    client->curdepth--;
  }
  // Casual treatment
  if (client->toaccess->NbElem()>0) {
    // Objects to access remain
    // Unstack in rootoid (next object to access)
    client->rootoid=client->toaccess->Scan(1);
    client->toaccess->Delete(1);
    // Accessed object becomes a root (last pos.)
    client->roots->Insert(client->roots->NbElem()+1,client->rootoid);
  } else {
    if (client->roots->NbElem()>0) { // Next level
      client->curdepth--;
      if (client->curdepth>=0) {
        // All elements in roots are unstacked
        n=client->roots->NbElem();
        for (i=0; i<n; i++) {
          k=client->roots->Scan(1);
          client->roots->Delete(1);
          // Objects pointed by this root are added into toaccess
          for (j=0; j<DB->GetObj(k)->ClassPtr()->MaxNRef(); j++)
            if (DB->GetObj(k)->GetORef(j)!=NULL)
              client->toaccess->Insert(client->toaccess->NbElem()+1,DB->GetObj(k)->GetORef(j)->OID());
        }
        // Access to 1st object in toaccess
        if (client->toaccess->NbElem()>0) {
          // Unstack in rootoid
          client->rootoid=client->toaccess->Scan(1);
          client->toaccess->Delete(1);
          // Accessed object becomes a root
          client->roots->Insert(client->roots->NbElem()+1,client->rootoid);
        } else client->curdepth=-1; // End
      }
    } else client->curdepth=-1; // End
  }
  if (client->curdepth<0) { // End of transaction: free memory
    delete client->roots;
    delete client->toaccess;
  }
}

// CLASS TransMgr: R2.2 - Simple traversal

void TransMgr::NextSimp(Client *client) {

  int i;
  if (client->curdepth==SIMDEPTH) { // 1st passage
    // Creation of list toaccess as a stack with rootoid on top
    client->toaccess=new ChList();
    client->toaccess->Insert(1,client->rootoid);
    // Creation of list roots => list of depths
    client->roots=new ChList();
    client->roots->Insert(1,SIMDEPTH-1);
  }
  // Casual treatment
  if (client->toaccess->NbElem()>0) {
    // Object to access is unstacked in rootoid
    client->rootoid=client->toaccess->Scan(1);
    client->toaccess->Delete(1);
    // Its depth in curdepth
    client->curdepth=client->roots->Scan(1);
    client->roots->Delete(1);
    // If depth not reached, stack the "child" objects
    if (client->curdepth>0) {
      for (i=DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()-1; i>=0; i--)
        if (DB->GetObj(client->rootoid)->GetORef(i)!=NULL) {
          client->toaccess->Insert(1,DB->GetObj(client->rootoid)->GetORef(i)->OID());
          client->roots->Insert(1,client->curdepth-1);
        }
    }
  } else { // End of traversal
    client->curdepth=-1;
    delete client->toaccess;
    delete client->roots;
  }
}

// CLASS TransMgr: R2.3 - Hierarchy traversal

void TransMgr::NextHier(Client *client) {

  int i;

  if (client->curdepth==HIEDEPTH) { // 1st passage
    // Creation of list toaccess as a stack with rootoid on top
    client->toaccess=new ChList();
    client->toaccess->Insert(1,client->rootoid);
    // Creation of list roots => list of depths
    client->roots=new ChList();
    client->roots->Insert(1,HIEDEPTH-1);
  }
  // Casual treatment
  if (client->toaccess->NbElem()>0) {
    // Object to access is unstacked in rootoid
    client->rootoid=client->toaccess->Scan(1);
    client->toaccess->Delete(1);
    // Its depth in curdepth
    client->curdepth=client->roots->Scan(1);
    client->roots->Delete(1);
    // If depth not reached, stack the "child" objects
    if (client->curdepth>0) {
      for (i=DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()-1; i>=0; i--)
        if ((DB->GetObj(client->rootoid)->GetORef(i)!=NULL)
        && (DB->GetObj(client->rootoid)->ClassPtr()->GetTRef(i)==client->reftype)) {
          client->toaccess->Insert(1,DB->GetObj(client->rootoid)->GetORef(i)->OID());
          client->roots->Insert(1,client->curdepth-1);
        }
    }
  } else { // End of traversal
    client->curdepth=-1;
    delete client->toaccess;
    delete client->roots;
  }
}

// CLASS TransMgr: R2.4 - Stochastic traversal

void TransMgr::NextStoc(Client *client) {

  int i;
  short k;
  float alea;

  client->curdepth--;
  if ((client->curdepth<(STODEPTH-1)) && (client->curdepth>=0)) { // Casual treatment
    // Random roll
    alea=Uni(0,1);
    // Test relationship to follow (i)
    i=0;
    k=-1;
    while ((i<DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()) && (k<0)) {
      if (alea>STOCHTAB[i]) i++;
      else k=1;
    }
    if (k==1) {// Relationship found
      if (DB->GetObj(client->rootoid)->GetORef(i)!=NULL) {
        client->rootoid=DB->GetObj(client->rootoid)->GetORef(i)->OID();
      } else client->curdepth=-1; // End of traversal
    } else client->curdepth=-1;
  }
}

// CLASS TransMgr: R6 - Transaction treatment

void TransMgr::TrTrans() {

  // INSERT CODE HERE for specific treatment
}

/////////////////////////////////////////////////////////////////////
// CLASS ClustMgr (AR3)
/////////////////////////////////////////////////////////////////////

// CLASS ClustMgr: Constructor

ClustMgr::ClustMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsb): Resource(n, cap, sim) {

  rsproc[0]=rsp;
  rsbd=rsb;
}

// CLASS ClustMgr: Initialize pointer toward BuffMgr

void ClustMgr::SetBuffPtr(BuffMgr *ptr) {

  buffmgr=ptr;
}

// CLASS ClustMgr: Initialize pointer toward ObjMgr

void ClustMgr::SetObjPtr(ObjMgr *ptr) {

  objmgr=ptr;
}

// CLASS ClustMgr: Event #0 - Arrival on AR3

void ClustMgr::Event30(Client *client) {

  if ((client->type>=1) && (client->type<=4))
    this->P(31,client,1);       // Reservation AR3 for casual treatment
  else if (client->type==5)
    this->P(34,client,1);       // Reservation AR3 for clustering
}

// CLASS ClustMgr: Event #1 - Tt: Wait for processor

void ClustMgr::Event31(Client *client) {

  rsproc[client->cliid]->P(32,client,1);
}

// CLASS ClustMgr: Event #2 - Tt: R7

void ClustMgr::Event32(Client *client) {

  float cltt_time;

  cltt_time=ClustTt(client);              // R7
  TCTIME+=cltt_time;
  Sim()->Sched()->Schedule(33,Sim()->Tnow()+cltt_time,client); // End of treatment
}

// CLASS ClustMgr: Event #3 - Tt: End of treatment

void ClustMgr::Event33(Client *client) {

  rsproc[client->cliid]->V();   // Release processor
  this->V();                    // Release clustering manager (AR3)
  Sim()->Sched()->Schedule(20,Sim()->Tnow(),client); // Transaction resumes
}


// CLASS ClustMgr: Event #4 - Cl: Lock database

void ClustMgr::Event34(Client *client) {

  Client *ficli;
  int i;

  // Creation of MULTILVL-1 fictitious clients to lock the DB
  for (i=0; i<MULTILVL-1; i++) {
    ficli=Sim()->NewClient();
    ficli->type=9;
    rsbd->P(35,ficli,9);
  }
  // DB lock end
  rsbd->P(36,client,8);
}

// CLASS ClustMgr: Event #5 - Cl: Destroy fictitious clients 

void ClustMgr::Event35(Client *client) {

  Sim()->KillClient(client);
}

// CLASS ClustMgr: Event #6 - Cl: Wait for processor

void ClustMgr::Event36(Client *client) {

  rsproc[client->cliid]->P(37,client,1);
}

// CLASS ClustMgr: Event #7 - Cl: R8

void ClustMgr::Event37(Client *client) {

  float clustering_time;

  clustering_time=Cluster(client);              // R8
  // printf("Clustering time=%f (end at time %f)\n",clustering_time,Sim()->Tnow()+clustering_time);
  CTIME+=clustering_time;
  Sim()->Sched()->Schedule(38,Sim()->Tnow()+clustering_time,client); // End of clustering
}

// CLASS ClustMgr: Event #8 - Cl: End of clustering

void ClustMgr::Event38(Client *client) {

  int i;

  rsproc[client->cliid]->V();   // Release processor
  for (i=0; i<MULTILVL; i++) rsbd->V(); // Release whole DB
  this->V();                    // Release AR3
  Sim()->KillClient(client);    // OUT
  NBCL++;
}

// CLASS ClustMgr: R7 - Perform treatment relative to clustering, return time needed
//                      DSTC strategy (Observation, Selection, and Consolidation phases)

float ClustMgr::ClustTt(Client *client) {

  int place, i, j, k, i1, i2, ifc, gflag;
  short stop;
  Client *newcli;
  float fe, time;

  // Observation phase

  time=0;

  // Update VO
  DB->vo[client->rootoid]++;
  // If there is a previously accessed object, update MO
  if (client->prevoid>=0) {
    // New entry
    DB->ni++;
    // Check if counter already exists
    place=1;
    stop=0;
    while ((place<=DB->mo[client->prevoid][0]->NbElem()) && (!stop))
      if (DB->mo[client->prevoid][0]->Scan(place)<client->rootoid) place++;
      else stop=1;
    if (stop) {
      if (DB->mo[client->prevoid][0]->Scan(place)==client->rootoid) {
        DB->mo[client->prevoid][1]->Update(place,DB->mo[client->prevoid][1]->Scan(place)+1);
      } else { // Counter to insert in position "place" (val=1)
        DB->mo[client->prevoid][0]->Insert(place,client->rootoid);
        DB->mo[client->prevoid][1]->Insert(place,1);
      }
    } else { // Counter to insert at queue (place)
      DB->mo[client->prevoid][0]->Insert(place,client->rootoid);
      DB->mo[client->prevoid][1]->Insert(place,1);
    }
  }
  // Previous object evolution
  client->prevoid=client->rootoid;
  // printf("DSTC: Observation entry #%d at time %f\n",DB->ni,Sim()->Tnow());

  // Selection and Consolidation phases

  if (DB->ni>=N) {
    // printf("DSTC: Selection + Consolidation at time %f\n",Sim()->Tnow());
    // Creation of clustering demand client 
    newcli=Sim()->NewClient();
    newcli->cliid=client->cliid;
    newcli->type=5;
    newcli->o1=new ChList;
    newcli->o2=new ChList;
    newcli->fc=new ChList;
    // Traverse MO
    for (i=0; i<NO; i++) {
      for (k=1; k<=DB->mo[i][0]->NbElem(); k++) {
        j=DB->mo[i][0]->Scan(k);
        if ((i!=j) && ((DB->vo[i]>=TFA) || (DB->vo[j]>=TFA))) { // Valid measurements
          // Compute MO(i,j)/VO(i)
          if (DB->vo[i]>0) fe=1.0*DB->mo[i][1]->Scan(k)/DB->vo[i];
          else fe=0;
          // Search for MO(j,i)
          place=1;
          stop=0;
          while ((place<=DB->mo[j][0]->NbElem()) && (!stop))
            if (DB->mo[j][0]->Scan(place)<i) place++;
            else stop=1;
          if (stop)
            if (DB->mo[j][0]->Scan(place)==i) {
              // Compute MO(j,i)/VO(j)
              if (DB->vo[j]>0) fe+=1.0*DB->mo[j][1]->Scan(place)/DB->vo[j];
              // Delete MO(j,i) not to compute the same thing again
              DB->mo[j][0]->Delete(place);
              DB->mo[j][1]->Delete(place);
            }
          // printf("fe(%d,%d)=%f\n",i,j,fe);
          if (fe>=TFE) { // Valid measurements for elementary matrix ME
            // MC is triangular => indices ordering
            if (i<j) {
              i1=i;
              i2=j;
            } else {
              i1=j;
              i2=i;
            }
            // Search for old fc in MC
            place=1;
            stop=0;
            while ((place<=DB->mc[i1][0]->NbElem()) && (!stop))
              if (DB->mc[i1][0]->Scan(place)<i2) place++;
              else stop=1;
            if (stop) {
              if (DB->mc[i1][0]->Scan(place)==i2) {
                // Old fc found => update
                ifc=(int)(W*DB->mc[i1][1]->Scan(place)+(1-W)*fe*100);
                DB->mc[i1][1]->Update(place,ifc);
                DB->mc[i1][2]->Update(place,DB->pi);
                gflag=DB->mc[i1][3]->Scan(place);
              } else {
                // New fc factor
                ifc=(int)(fe*100);
                gflag=0;
                DB->mc[i1][0]->Insert(place,i2);
                DB->mc[i1][1]->Insert(place,ifc);
                DB->mc[i1][2]->Insert(place,DB->pi);
                DB->mc[i1][3]->Insert(place,gflag);
              }
            } else {
              // New fc factor
              ifc=(int)(fe*100);
              gflag=0;
              DB->mc[i1][0]->Insert(place,i2);
              DB->mc[i1][1]->Insert(place,ifc);
              DB->mc[i1][2]->Insert(place,DB->pi);
              DB->mc[i1][3]->Insert(place,gflag);
            }
            // printf("ifc=%d, TFC*100=%d, gflag=%d\n",ifc,(int)(TFC*100),gflag);
            if ((ifc>=TFC*100) && (!gflag)) {
              // Update clustering flag
              place=1;
              stop=0;
              while ((place<=DB->mc[i1][0]->NbElem()) && (!stop))
                if (DB->mc[i1][0]->Scan(place)<i2) place++;
                else stop=1;
              if (stop)
                if (DB->mc[i1][0]->Scan(place)==i2)
                  DB->mc[i1][3]->Update(place,1);
              // Add couple (i1, i2) into list of objects to cluster
              newcli->o1->Insert(1,i1);
              newcli->o2->Insert(1,i2);
              newcli->fc->Insert(1,ifc);
            }
          }
        }
      }
    }
//    DB->DisplayMO();
//    DB->DisplayMC();
    DB->NewPeriod();
    // Trigger clustering
    if (newcli->o1->NbElem()>0) {
//      printf("DSTC clustering triggered at time %f\n",Sim()->Tnow());
      Sim()->Sched()->Schedule(30,Sim()->Tnow(),newcli);
      }
    else {
      delete newcli->o1;
      delete newcli->o2;
      delete newcli->fc;
      Sim()->KillClient(newcli);
    }
  }

  return time;
}

// CLASS ClustMgr: R8 - Perform clustering, return time needed

float ClustMgr::Cluster(Client *client) {

  lastpage=-999;
  if (CLUSTP==9) { // User-deENDed clustering policy
    return DSTC(client);
  } else return 0; // Else no action
}

// CLASS ClustMgr: Clustering by the DSTC algorithm, return time needed

float ClustMgr::DSTC(Client *client) {

  ChList *ur[3];                        // (re)Clustering unit (UR)
                                        // 0 - OID object #1
                                        // 1 - OID object #2
                                        // 2 - Weight (fc)
  ChList *cluster;                      // Current cluster 
  int i, j, k, n, place, place2, rsize, v1, v2, v3, v4, obj;
  short stop, stop2;
  Page *newpage, *cpage;
  float time;

  time=0;

  //printf("DSTC! %d\n",client->o1->NbElem());

  // Enrich (o1,o2) list from MC

  n=client->o1->NbElem();
  place=n+1;
  // Traverse current (o1,o2) list
  for (i=1; i<=n; i++) {
    // o1
    v1=client->o1->Scan(i);
    // Traverse MC for current object v1
    for (j=1; j<=DB->mc[v1][0]->NbElem(); j++) {
      v2=DB->mc[v1][0]->Scan(j);
      if (DB->mc[v1][1]->Scan(j)>=TFC) {
        // Traverse (o1,o2) to check if the couple (v1,v2) is already in it
        k=1;
        stop=0;
        while ((k<=ur[0]->NbElem()) && (!stop)) {
          v3=client->o1->Scan(k);
          v4=client->o2->Scan(k);
          if ( ((v1==v3) && (v2==v4)) || ((v1==v4) && (v2==v3)) ) stop=1;
          else k++;
        }
        //printf("END while ((k<=ur[0]->NbElem()) && (!stop)) j=%d\n",j);
        if (!stop) {
          client->o1->Insert(place,v1);
          client->o2->Insert(place,v2);
          client->fc->Insert(place,DB->mc[v1][1]->Scan(j));
          place++;
        }
      }
    }
    //printf("END for (j=1; j<=DB->mc[v1][0]->NbElem(); j++) i=%d\n",i);
    // o2
    v1=client->o2->Scan(i);
    // Traverse MC for current object v1
    for (j=1; j<=DB->mc[v1][0]->NbElem(); j++) {
      v2=DB->mc[v1][0]->Scan(j);
      if (DB->mc[v1][1]->Scan(j)>=TFC) {
        // Traverse (o1,o2) to check if the couple (v1,v2) is already in it
        k=1;
        stop=0;
        while ((k<=ur[0]->NbElem()) && (!stop)) {
          v3=client->o1->Scan(k);
          v4=client->o2->Scan(k);
          if ( ((v1==v3) && (v2==v4)) || ((v1==v4) && (v2==v3)) ) stop=1;
          else k++;
        }
        //printf("END while ((k<=ur[0]->NbElem()) && (!stop)) j=%d\n",j);
        if (!stop) {
          client->o1->Insert(place,v1);
          client->o2->Insert(place,v2);
          client->fc->Insert(place,DB->mc[v1][1]->Scan(j));
          place++;
        }
      }
    }
    //printf("END for (j=1; j<=DB->mc[v1][0]->NbElem(); j++) i=%d\n",i);
  }
  //printf("END for (i=1; i<=n; i++)\n");
/*
  printf("o1 ");client->o1->Display();
  printf("o2 ");client->o2->Display();
  printf("fc ");client->fc->Display();  */

  // UR building

  while (client->o1->NbElem()>0) {
    // Current UR creation (list of OIDs)
    ur[0]=new ChList;
    ur[1]=new ChList;
    ur[2]=new ChList;
    // 1st couple (o1,o2) "unstacked" in current UR 
    ur[0]->Insert(1,client->o1->Scan(1));
    client->o1->Delete(1);
    ur[1]->Insert(1,client->o2->Scan(1));
    client->o2->Delete(1);
    ur[2]->Insert(1,client->fc->Scan(1));
    client->fc->Delete(1);
    // Enrich current UR from (o1,o2)
    stop=0;
    while ((!stop) && (client->o1->NbElem()>0)) {
      // Traverse (o1,o2)
      stop=1;
      place=1;
      while (place<=client->o1->NbElem()) {
        v1=client->o1->Scan(place);
        v2=client->o2->Scan(place);
        // printf("(o1,o2)=(%d,%d)\n",v1,v2);
        // Traverse current UR to find connex components
        stop2=0;
        place2=1;
        while ((place2<=ur[0]->NbElem()) && (!stop2)) {
          v3=ur[0]->Scan(place2);
          v4=ur[1]->Scan(place2);
          // printf("(ur1,ur2)=(%d,%d)\n",v3,v4);
          if ((v1==v3) || (v2==v3) || (v1==v4) || (v2==v4)) {
            // Connex component detected
            stop2=1;
            stop=0;
            // Insertion into current UR 
            ur[0]->Insert(1,v1);
            ur[1]->Insert(1,v2);
            ur[2]->Insert(1,client->fc->Scan(place));
            // Couple deleted from (o1,o2)
            client->o1->Delete(place);
            client->o2->Delete(place);
            client->fc->Delete(place);
            place--;
          } else place2++;
        }
        place++;
      }
    } // END while ((!stop) && (client->o1->NbElem()>0)) {

/*
    printf("Current UR: \n");
    ur[0]->Display();
    ur[1]->Display();
    ur[2]->Display();   */

    // Current UR ordering, output = cluster

    cluster=Schedule(ur);
//    printf("Current cluster ");cluster->Display();
    NCLUSTERS++;
    TCLUSTERS+=cluster->NbElem();

    // Placement of cluster on disk

    // Create new page
    newpage=DB->NewPage();
    DB->AddPage(newpage);
    // Placement of all the objects from the cluster
    for (i=1; i<=cluster->NbElem(); i++) {
      obj=cluster->Scan(i);
      // Load current object
      client->rootoid=obj;
      time+=ObjRequest(client);
      // Current object deleted from the page(s) where it's stored
//      printf("Placement of object #%d\n",obj);
      while (DB->GetObj(obj)->PageDisk()->NbElem()>0) {
        // Unstacking
        v1=DB->GetObj(obj)->PageDisk()->Scan(1);
        cpage=DB->GetPage(v1);
        DB->GetObj(obj)->PageDisk()->Delete(1);
        // Delete object in page cpage/v1
        if (cpage!=NULL) {
          rsize=cpage->Free();
          j=1;
          stop=0;
          while ((j<=cpage->Objects()->NbElem()) && (!stop))
            if (cpage->Objects()->Scan(j)==obj) stop=1;
            else j++;
          if (stop) {
            // Deletion
            cpage->Objects()->Delete(j);
            // Evolution of remaining size
            rsize+=DB->GetObj(obj)->ClassPtr()->GetSize();
          }
//          printf(" Object #%d removed from page #%d, free space=%d\n",obj,v1,rsize);
          if (rsize<PGSIZE) // Update free space
            cpage->ResetFree(rsize);
          else // Empty page => deleted
            DB->DelPage(v1);
        } // END if (cpage!=NULL)
        // Page v1 update flag set to 1 in the buffer
        buffmgr->SetUpdated(v1);
      } // END while (DB->GetObj(obj)->PageDisk()->NbElem()>0)
      // Placement of current object
      if (DB->GetObj(obj)->ClassPtr()->GetSize()<=PGSIZE) {
        // Normal object with size smaller than the page size
        if (DB->GetObj(obj)->ClassPtr()->GetSize()>newpage->Free()) { // Taille > place libre
          // Save previous page
          time+=AccDisk(newpage->PageNum());
          // New page
          newpage=DB->NewPage();
          DB->AddPage(newpage);
        }
        // Object inserted in page content
        newpage->Objects()->Insert(1,obj);
        // Page inserted in object page list
        DB->GetObj(obj)->PageDisk()->Insert(1,newpage->PageNum());
        // Update free space
        newpage->ResetFree(newpage->Free()-DB->GetObj(obj)->ClassPtr()->GetSize());
//        printf("Object #%d moved to page #%d\n",obj,newpage->PageNum());
      } else { // Large objet
        // rsize: size remaining to place
        rsize=DB->GetObj(obj)->ClassPtr()->GetSize()-newpage->Free();
        // Current page full
        newpage->ResetFree(0);
        // Object inserted in page content
        newpage->Objects()->Insert(1,obj);
        // Page inserted in object page list
        DB->GetObj(obj)->PageDisk()->Insert(1,newpage->PageNum());
//        printf("Object #%d moved to page #%d\n",obj,newpage->PageNum());
        // While object is not wholly stored
        while (rsize>0) {
          // New page
          newpage=DB->NewPage();
          DB->AddPage(newpage);
          // Object inserted in page content
          newpage->Objects()->Insert(1,obj);
          // Page inserted in object page list
          DB->GetObj(obj)->PageDisk()->Insert(1,newpage->PageNum());
          if (newpage->Free()<rsize) {   // Object does not fit
            rsize=rsize-newpage->Free(); // Remaining size --
            newpage->ResetFree(0);       // Page full
          } else {
            newpage->ResetFree(newpage->Free()-rsize); // Free space --
            rsize=0;
          }
//          printf("Object #%d moved to page #%d\n",obj,newpage->PageNum());
        }
      }
    } // END for (i=1; i<=cluster->NbElem();
    // Save last page
    time+=AccDisk(newpage->PageNum());

    // Destroy current UR & current cluster

    delete ur[0];
    delete ur[1];
    delete ur[2];
    delete cluster;

  } // END while (client->o1->NbElem()>0)

  // Destroy clustering lists

  delete client->o1;
  delete client->o2;
  delete client->fc;

  return time;

} // END of procedure

// CLASS ClustMgr: Ordering of UR into clusters (DSTC)

ChList *ClustMgr::Schedule(ChList *ur[3]) {

  Node *hnode;                  // Node list head
  int nodes;                    // Current number of nodes
  ChList *links[3];             // List of links
                                // 0 - Node #1
                                // 1 - Node #2
                                // 2 - Weight
  ChList *cluster;              // Output cluster
  int i, j, n, n1, n2, n3, n4;
  Node *cnode, *newnode, *prevnode;
  short found1, found2;
  int w, maxw, place;
  Node *pn1, *prevn1, *pn2, *prevn2;
  int t1, t2, q1, q2, wt1t2, wt1q2, wq1t2, wq1q2, w1, w2;
  int ndel, nkeep;

  // Creation of list of nodes and list of links

  hnode=NULL;
  nodes=0;
  links[0]=new ChList;
  links[1]=new ChList;
  links[2]=new ChList;
  for (i=1; i<=ur[0]->NbElem(); i++) {
    // Traverse list of nodes not to add the same object twice
    cnode=hnode;
    found1=0;
    found2=0;
    while (cnode!=NULL) {
      if (cnode->list->Scan(1)==ur[0]->Scan(i)) {
        found1=1;
        n1=cnode->num;
      }
      if (cnode->list->Scan(1)==ur[1]->Scan(i)) {
        found2=1;
        n2=cnode->num;
      }
      cnode=cnode->next;
    } // END while (cnode!=NULL)
    if (!found1) {                      // New node
      nodes++;
      newnode=new Node;
      newnode->num=nodes;
      newnode->list=new ChList;
      newnode->list->Insert(1,ur[0]->Scan(i));
      newnode->next=hnode;
      hnode=newnode;
      n1=newnode->num;
    }
    if (!found2) {                      // New node
      nodes++;
      newnode=new Node;
      newnode->num=nodes;
      newnode->list=new ChList;
      newnode->list->Insert(1,ur[1]->Scan(i));
      newnode->next=hnode;
      hnode=newnode;
      n2=newnode->num;
    }
    // Insertion current link into list of links
    links[0]->Insert(1,n1);
    links[1]->Insert(1,n2);
    links[2]->Insert(1,ur[2]->Scan(i));
  } // END for (i=1; i<=ur[0]->NbElem();


/*  printf("\nNodes list:\n");
  cnode=hnode;
  while (cnode!=NULL) {
    printf("%d)   ",cnode->num);
    cnode->list->Display();
    cnode=cnode->next;
  }
  printf("\nLinks list:\n");
  links[0]->Display();
  links[1]->Display();
  links[2]->Display();  */

  // Ordering, reduction to one single supernode

  while ((nodes>1)&&(links[0]->NbElem())) {
    // Search for the greatest weight in the list of links
    maxw=-1;
    for (i=1; i<=links[0]->NbElem(); i++) {
      w=links[2]->Scan(i);
      if (w>maxw) maxw=w;
    }
    // Search for links of maximum weight
    i=1;
    n=links[0]->NbElem();
    while (i<=n) {
      if (links[2]->Scan(i)>=maxw) { // Fusion of the nodes
        // Search for the nodes and their "head" and "queue" objects
        // head -> ti, queue -> qi
        cnode=hnode;
        prevnode=NULL;
        while (cnode!=NULL) {
          if (cnode->num==links[0]->Scan(i)) {
            pn1=cnode;
            prevn1=prevnode;
            t1=cnode->list->Scan(1);
            q1=cnode->list->Scan(cnode->list->NbElem());
          }
          if (cnode->num==links[1]->Scan(i)) {
            pn2=cnode;
            prevn2=prevnode;
            t2=cnode->list->Scan(1);
            q2=cnode->list->Scan(cnode->list->NbElem());
          }
          prevnode=cnode;
          cnode=cnode->next;
        } // END while (cnode!=NULL)
        // printf("t1=%d q1=%d t2=%d q2=%d\n",t1,q1,t2,q2);
        // Search in UR for link weights between all
        // the possible couples tt, tq, qq
        wt1t2=0;
        wt1q2=0;
        wq1t2=0;
        wq1q2=0;
        for (j=1; j<=ur[0]->NbElem(); j++) {
          w1=ur[0]->Scan(j);
          w2=ur[1]->Scan(j);
          //printf("w1=%d w2=%d Bond strength=%d\n",w1,w2,ur[2]->Scan(j));
          if ( ((w1==t1) && (w2==t2)) || ((w2==t1) && (w1==t2)) )
            wt1t2=ur[2]->Scan(j);
          if ( ((w1==t1) && (w2==q2)) || ((w2==t1) && (w1==q2)) )
            wt1q2=ur[2]->Scan(j);
          if ( ((w1==q1) && (w2==t2)) || ((w2==q1) && (w1==t2)) )
            wq1t2=ur[2]->Scan(j);
          if ( ((w1==q1) && (w2==q2)) || ((w2==q1) && (w1==q2)) )
            wq1q2=ur[2]->Scan(j);
        } // END for (j=1; j<=ur[0]->NbElem();
        //printf("wt1t2=%d wt1q2=%d wq1t2=%d wq1q2=%d\n",wt1t2,wt1q2,wq1t2,wq1q2);
        // Node fusion, respecting the rule
        if ( (wt1t2>=wt1q2) && (wt1t2>=wq1t2) && (wt1t2>=wq1q2) ) { // t1 t2
          //printf("Fusion t1t2\n");
          // Save number of node to delete and number of node to keep
          ndel=links[0]->Scan(i);
          nkeep=links[1]->Scan(i);
          // Fusion
          pn2->list->Reverse();
          place=pn2->list->NbElem()+1;
          for (j=1; j<=pn1->list->NbElem(); j++) {
            pn2->list->Insert(place,pn1->list->Scan(j));
            place++;
          }
          // Destroy node pn1
          if (prevn1!=NULL) prevn1->next=pn1->next;
          else hnode=pn1->next;
          delete pn1->list;
          delete pn1;
          nodes--;
        } else if ( (wt1q2>=wt1t2) && (wt1q2>=wq1t2) && (wt1q2>=wq1q2) ) { // t1 q2
          //printf("Fusion t1q2\n");
          // Save number of node to delete and number of node to keep
          ndel=links[0]->Scan(i);
          nkeep=links[1]->Scan(i);
          // Fusion
          place=pn2->list->NbElem()+1;
          for (j=1; j<=pn1->list->NbElem(); j++) {
            pn2->list->Insert(place,pn1->list->Scan(j));
            place++;
          }
          // Destroy node pn1
          if (prevn1!=NULL) prevn1->next=pn1->next;
          else hnode=pn1->next;
          delete pn1->list;
          delete pn1;
          nodes--;
        } else if ( (wq1t2>=wt1t2) && (wq1t2>=wt1q2) && (wq1t2>=wq1q2) ) { // q1 t2
          //printf("Fusion q1t2\n");
          // Save number of node to delete and number of node to keep
          ndel=links[1]->Scan(i);
          nkeep=links[0]->Scan(i);
          // Fusion
          place=pn1->list->NbElem()+1;
          for (j=1; j<=pn2->list->NbElem(); j++) {
            pn1->list->Insert(place,pn2->list->Scan(j));
            place++;
          }
          // Destroy node pn2
          if (prevn2!=NULL) prevn2->next=pn2->next;
          else hnode=pn2->next;
          delete pn2->list;
          delete pn2;
          nodes--;
        } else { // q1 q2
          //printf("Fusion q1q2\n");
          // Save number of node to delete and number of node to keep
          ndel=links[1]->Scan(i);
          nkeep=links[0]->Scan(i);
          // Fusion
          pn2->list->Reverse();
          place=pn1->list->NbElem()+1;
          for (j=1; j<=pn2->list->NbElem(); j++) {
            pn1->list->Insert(place,pn2->list->Scan(j));
            place++;
          }
          // Destroy node pn2
          if (prevn2!=NULL) prevn2->next=pn2->next;
          else hnode=pn2->next;
          delete pn2->list;
          delete pn2;
          nodes--;
        }
        // Destroy current link
        links[0]->Delete(i);
        links[1]->Delete(i);
        links[2]->Delete(i);
        i--;

        // Update links containing the deleted node

        for (j=1; j<=links[0]->NbElem(); j++) {
          if (links[0]->Scan(j)==ndel) links[0]->Update(j,nkeep);
          if (links[1]->Scan(j)==ndel) links[1]->Update(j,nkeep);
        }

/*      More time-costly solution 
        int m;
        j=1;
        m=links[0]->NbElem();
        while (j<=m) {
          if (links[0]->Scan(j)==ndel) links[0]->Update(j,nkeep);
          if (links[1]->Scan(j)==ndel) links[1]->Update(j,nkeep);
          if (links[0]->Scan(j)==links[1]->Scan(j)) {
            links[0]->Delete(j);
            links[1]->Delete(j);
            links[2]->Delete(j);
            if (j<=i) i--;
            j--;
          }
          m=links[0]->NbElem();
          j++;
        }     */


/*      printf("\nNodes list after fusion:\n");
      cnode=hnode;
      while (cnode!=NULL) {
        printf("%d)   ",cnode->num);
        cnode->list->Display();
        cnode=cnode->next;
      }
      printf("\nLinks list after fusion:\n");
      links[0]->Display();
      links[1]->Display();
      links[2]->Display();      */

      //////
      i=n+1; // STOP
      //////
      } // END if (links[3]->Scan(i)>=wmax)
      i++;
      n=links[0]->NbElem();
    } // END while (i<=n)
    // Elimination of redundant links
    i=1;
    n=links[0]->NbElem();
    while (i<=n) {
      n1=links[0]->Scan(i);
      n2=links[1]->Scan(i);
      w1=links[2]->Scan(i);
      j=i+1;
      while (j<=n) {
        n3=links[0]->Scan(j);
        n4=links[1]->Scan(j);
        if ( ((n1==n3) && (n2==n4)) || ((n1==n4) && (n2==n3)) ) {
          w2=links[2]->Scan(j);
          // Delete link with lower weight
          if (w1<w2) {
            links[0]->Delete(i);
            links[1]->Delete(i);
            links[2]->Delete(i);
            i--;
            j=n+1;
          } else {
            links[0]->Delete(j);
            links[1]->Delete(j);
            links[2]->Delete(j);
            j--;
            i++;
          }
        } // END big if
        j++;
        n=links[0]->NbElem();
      } // END while (j<=n)
      i++;
      n=links[0]->NbElem();
    } // END while (i<=n)

/*    printf("\nConsistent links list:\n");
    links[0]->Display();
    links[1]->Display();
    links[2]->Display();
    printf("Nodes remaining: %d\n",nodes);      */
  } // END while (nodes>1)

  // Copy final supernode into cluster

  cluster=new ChList;
  for (i=1; i<=hnode->list->NbElem(); i++)
    cluster->Insert(1,hnode->list->Scan(i));

  // Destroy list of nodes
  // (head of list points toward the last supernode)

  delete hnode->list;
  delete hnode;

  // Destroy list of links

  delete links[0];
  delete links[1];
  delete links[2];

  // Return cluster pointer

  return cluster;

} // END of procedure

// CLASS ClustMgr: Object request, return time needed

float ClustMgr::ObjRequest(Client *client) {

  float time;
  int i, n;

  // R3: extract pages containing object client->rootoid
  objmgr->ExtPg(client);
  // Prefetching (enrich client pgnums list)
  objmgr->Prefetch(client);

  // R4: Check pages in memory
  buffmgr->AccPage(client);
  // Disk access on pages to load
  for (i=1; i<=client->pgnums->NbElem(); i++)
    time+=AccDisk(client->pgnums->Scan(i));
  // Insert all accessed pages into the buffer
  for (i=1; i<=client->pgnums->NbElem(); i++) {
    n=buffmgr->Replace();             // Find a position (n)
    if (buffmgr->Updated(n)) // Updated pages => write
      time+=AccDisk(buffmgr->Page(n));
    // New page is inserted in the buffer at position n
    buffmgr->SetBuffer(n,client->pgnums->Scan(i),Sim()->Tnow()+time);
  }
  delete client->pgnums;        // Free memory
  time+=buffmgr->Tnet();        // Network transfer BuffMgr->ObjMgr

  // Network transfer ObjMgr -> ClustMgr
  time+=objmgr->Tnet()*DB->GetObj(client->rootoid)->ClassPtr()->GetSize();

  return time;
}

// CLASS ClustMgr: Direct disk access, return time needed

float ClustMgr::AccDisk(int pagenum) {

  float time;

  time=DISKLAT+DISKTRA;
  if ((pagenum<(lastpage-1)) || (pagenum>(lastpage+1))) time+=DISKSEA; // Non-contiguous access
  lastpage=pagenum;
  NBIO++;
  CNBIO++;
  return time;
}

/////////////////////////////////////////////////////////////////////
// CLASS ObjMgr (AR4)
/////////////////////////////////////////////////////////////////////

// CLASS ObjMCl: Constructor

ObjMgr::ObjMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, float t): Resource(n, cap, sim) {

  rsproc[0]=rsp;
  tnet=t;
}

// CLASS ObjMCl: Return tnet

float ObjMgr::Tnet() {

  return tnet;
}

// CLASS ObjMCl: Event #0 - Reservation AR4

void ObjMgr::Event40(Client *client) {

  this->P(41,client,1);
}

// CLASS ObjMCl: Event #1 - Wait for processor

void ObjMgr::Event41(Client *client) {

  rsproc[client->cliid]->P(42,client,1);
}

// CLASS ObjMCl: Event #2 - R3 + prefetching

void ObjMgr::Event42(Client *client) {

  // R3
  ExtPg(client);
  // Prefetching (enrich client pgnums list)
  Prefetch(client);
  // Release processor
  rsproc[client->cliid]->V();
  // Page request
  Sim()->Sched()->Schedule(50,Sim()->Tnow(),client); // Request transmitted
}

// CLASS ObjMCl: Event #3 - Object in memory

void ObjMgr::Event43(Client *client) {

  this->V();                            // Release AR4
  Sim()->Sched()->Schedule(26,Sim()->Tnow()+tnet*DB->GetObj(client->rootoid)->ClassPtr()->GetSize(),client);
  // Stat number of accessed objects
  NBOBJ[client->type-1]++;
  NBOBJ[NTTRANS]++;
}

// CLASS ObjMCl: R3 - Extract page

void ObjMgr::ExtPg(Client *client) {

  int i, oid;

  client->pgnums=new ChList();
  for (i=1; i<=DB->GetObj(client->rootoid)->PageDisk()->NbElem(); i++) {
    oid=DB->GetObj(client->rootoid)->PageDisk()->Scan(i);
    client->pgnums->Insert(1,oid);
  }

//  printf("Object #%d page ",client->rootoid);
//  client->pgnums->Display();
}

// CLASS ObjMCl: Prefetching routine 

void ObjMgr::Prefetch(Client *client) {

  if (PREFETCH==9) {                    // User-defined strategy
    // Insert code here
    // Enrich client->pgnums list with the page to access
  }
  // Else no treatment
//  printf("[After prefetching] Object #%d page ",client->rootoid);
//  client->pgnums->Display();
}

/////////////////////////////////////////////////////////////////////
// CLASS BuffMgr (AR5)
/////////////////////////////////////////////////////////////////////

// CLASS BuffMgr: Constructor

BuffMgr::BuffMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, float t): Resource(n, cap, sim) {

  rsproc=rsp;
  tnet=t;
}

// CLASS BuffMgr: Return tnet

float BuffMgr::Tnet() {

  return tnet;
}

// CLASS BuffMgr: Return 1 if page #num was "updated", 0 otherwise

short BuffMgr::Updated(int num) {

  if (num<=buffpos) return buffer[num].updated;
  else return 0;
}

// CLASS BuffMgr: Return number of page in position num

int BuffMgr::Page(int num) {

  if (num<=buffpos) return buffer[num].pgnum;
  else return -1;
}

// CLASS BuffMgr: Buffer initialization to empty

void BuffMgr::InitBuffer() {

  int i, j;

  for (i=0; i<BUFFSIZE; i++) {
    buffer[i].pgnum=-1;
    buffer[i].updated=0;
    buffer[i].nbacc=0;
    for (j=0; j<KLRU; j++) buffer[i].date[j]=0;
    buffer[i].clockflag=0;
  }
  buffpos=-1;
  clock=0;
}

// CLASS BuffMgr: Mark page #num to "updated"

void BuffMgr::SetUpdated(int num) {

  int i;

  i=0;
  while (i<=buffpos) {
    if (buffer[i].pgnum==num) {
      buffer[i].updated=1;
      i=buffpos+1;
    } else i++;
  }
}

// CLASS BuffMgr: Insert page #pgnum in position num

void BuffMgr::SetBuffer(int num,int pgnum,float time) {

  int j;

  buffer[num].pgnum=pgnum;
  buffer[num].updated=0;
  buffer[num].nbacc=1;
  for (j=0; j<KLRU; j++) buffer[num].date[j]=time;
  buffer[num].clockflag=1;
}

// CLASS BuffMgr: Event #0 - Reservation AR5

void BuffMgr::Event50(Client *client) {

  this->P(51,client,1);
}

// CLASS BuffMgr: Event #1 - Wait for processor (1)

void BuffMgr::Event51(Client *client) {

  rsproc->P(52,client,1);
}

// CLASS BuffMgr: Event #2 - R4 and I/O request if necessary

void BuffMgr::Event52(Client *client) {

  AccPage(client);              // R4
  rsproc->V();                  // Release processor
  // Remaining pages in pgnums must be loaded
  if (client->pgnums->NbElem())
    Sim()->Sched()->Schedule(60,Sim()->Tnow(),client);
  else  // Nothing to load
    Sim()->Sched()->Schedule(55,Sim()->Tnow(),client);
}

// CLASS BuffMgr: Event #3 - Wait for processor (2)

void BuffMgr::Event53(Client *client) {

  rsproc->P(54,client,1);
}

// CLASS BuffMgr: Event #4 - Page replacement

void BuffMgr::Event54(Client *client) {

  int i, j, n;
  Client *newcli;

  // Insertion of all accessed pages into the buffer
  for (i=1; i<=client->pgnums->NbElem(); i++) {
    n=Replace();             // Find a slot (n)
    if (buffer[n].updated) { // Page updated => write it
      newcli=Sim()->NewClient();
      newcli->type=7;
      newcli->pgnums=new ChList;
      newcli->pgnums->Insert(1,buffer[n].pgnum);
      Sim()->Sched()->Schedule(60,Sim()->Tnow(),newcli); // Write request sent
    }
    // New page is inserted in the buffer at position n
    buffer[n].pgnum=client->pgnums->Scan(i);
    buffer[n].updated=0;
    buffer[n].nbacc=1;
    for (j=0; j<KLRU; j++) buffer[n].date[j]=Sim()->Tnow();
    buffer[n].clockflag=1;
  }
  rsproc->V();                  // Release processor
  Sim()->Sched()->Schedule(55,Sim()->Tnow(),client);
}

// CLASS BuffMgr: Event #5 - Network transfer

void BuffMgr::Event55(Client *client) {

  this->V();                    // Release AR5
  delete client->pgnums;        // Free memory
  Sim()->Sched()->Schedule(43,Sim()->Tnow()+tnet,client);
}

// CLASS BuffMgr: R4 - Access page(s)

// In output, the list of pages to access only contains the
// pages that are not in the buffer (pages to access on disk)

void BuffMgr::AccPage(Client *client) {

  int i, j, k, n;
  short paginmem;

  i=1;
  n=client->pgnums->Scan(i);
  while (n!=-1) {
    j=0;
    paginmem=0;
    while ((j<BUFFSIZE) && (!paginmem))
      if (buffer[j].pgnum==n) {
        paginmem=1;
//        printf("Object #%d: page #%d in memory\n",client->rootoid,n);
      } else j++;
    if (paginmem) {
      // Page in buffer
      client->pgnums->Delete(i);        // Deleted from pgnums
      // Replacement flags set-up
      buffer[j].nbacc++;
      if (KLRU>1)
        for (k=KLRU-1; k>=1; k--)
          buffer[j].date[k]=buffer[j].date[k-1];
      buffer[j].date[0]=Sim()->Tnow();
      buffer[j].clockflag=1;
    } else i++;
    n=client->pgnums->Scan(i);
  }
}

// CLASS BuffMgr: Buffer page replacement

int BuffMgr::Replace() {

  if (buffpos<BUFFSIZE-1) {     // Buffer not full, no replacement
    buffpos++;
    return buffpos;
  } else {                      // Buffer full => page to replace
    switch (PGREP) {
      case 1: return RANDOMrep();break;
      case 2: return FIFOrep();break;
      case 3: return LFUrep();break;
      case 4: return LRUrep();break;
      case 5: return CLOCKrep();break;
      case 6: return GCLOCKrep();break;
      case 9: return Otherep();break;
      default: printf("\nError: (%d) unknown page replacement policy!\n",PGREP);return -1;
    }
  }
}

// CLASS BuffMgr: Buffer page replacement policy - RANDOM

int BuffMgr::RANDOMrep() {

  return CRand(1,0,BUFFSIZE-1);
}

// CLASS BuffMgr: Buffer page replacement policy - FIFO

int BuffMgr::FIFOrep() {

  BuffSlot temp;
  int i, j;

  // Shift buffer slots one position to the "left". 1st slot is placed last
  // and will be replaced.

  temp.pgnum=buffer[0].pgnum;
  temp.updated=buffer[0].updated;
  temp.nbacc=buffer[0].nbacc;
  for (j=0; j<KLRU; j++) temp.date[j]=buffer[0].date[j];
  temp.clockflag=buffer[0].clockflag;

  for (i=0; i<BUFFSIZE-1; i++) {
    buffer[i].pgnum=buffer[i+1].pgnum;
    buffer[i].updated=buffer[i+1].updated;
    buffer[i].nbacc=buffer[i+1].nbacc;
    for (j=0; j<KLRU; j++) buffer[i].date[j]=buffer[i+1].date[j];
    buffer[i].clockflag=buffer[i+1].clockflag;
  }

  buffer[BUFFSIZE-1].pgnum=temp.pgnum;
  buffer[BUFFSIZE-1].updated=temp.updated;
  buffer[BUFFSIZE-1].nbacc=temp.nbacc;
  for (j=0; j<KLRU; j++) buffer[BUFFSIZE-1].date[j]=temp.date[j];
  buffer[BUFFSIZE-1].clockflag=temp.clockflag;

  return (BUFFSIZE-1);
}

// CLASS BuffMgr: Buffer page replacement policy - LFU

int BuffMgr::LFUrep() {

  int i, min, p;

  // Buffer traversal to find the page with the lowest number of accesses

  min=buffer[clock].nbacc;
  p=clock;
  i=clock+1;
  if (i>=BUFFSIZE) i=0;
  while (i!=clock) {
    if (buffer[i].nbacc<min) {
      min=buffer[i].nbacc;
      p=i;
    }
    i++;
    if (i>=BUFFSIZE) i=0;
  }
  clock=p+1;
  if (clock>=BUFFSIZE) clock=0;
  return p;
}

// CLASS BuffMgr: Buffer page replacement policy - LRU-K

int BuffMgr::LRUrep() {

  int i, j, p;
  float dist, max;

  // Buffer traversal to find the page with the biggest usage distance

  max=0;
  for (j=0; j<KLRU; j++) max+=Sim()->Tnow()-buffer[clock].date[j];
  p=clock;
  i=clock+1;
  if (i>=BUFFSIZE) i=0;
  while (i!=clock) {
    dist=0;
    for (j=0; j<KLRU; j++) dist+=Sim()->Tnow()-buffer[i].date[j];
    if (dist>max) {
      max=dist;
      p=i;
    }
    i++;
    if (i>=BUFFSIZE) i=0;
  }
  clock=p+1;
  if (clock>=BUFFSIZE) clock=0;
  return p;
}

// CLASS BuffMgr: Buffer page replacement policy - CLOCK

int BuffMgr::CLOCKrep() {

  // Buffer traversal until a page with flag set to 0 is found
  // (unused page). Flags set to 1 are reset to 0 on the way.

  while (buffer[clock].clockflag) {
    buffer[clock].clockflag=0;
    clock++;
    if (clock>=BUFFSIZE) clock=0;
  }
  return clock;
}

// CLASS BuffMgr: Buffer page replacement policy - GCLOCK

int BuffMgr::GCLOCKrep() {

  // Traverse buffer until a page with 0 number of accesses is found.
  // Decrease the number of accesses by 1 on the way.

  while (buffer[clock].nbacc) {
    buffer[clock].nbacc--;
    clock++;
    if (clock>=BUFFSIZE) clock=0;
  }
  return clock;
}

// CLASS BuffMgr: Buffer page replacement policy - User-defined policy

int BuffMgr::Otherep() {

  // Insert code here
}

/////////////////////////////////////////////////////////////////////
// CLASS IOSubSys (AR6)
/////////////////////////////////////////////////////////////////////

// CLASS IOSubSys: Constructor

IOSubSys::IOSubSys(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsd): Resource(n, cap, sim) {

  rsproc=rsp;
  rsdisk=rsd;
  lastpage=-999;
}

// CLASS IOSubSys: Event #0 - Reservation AR6

void IOSubSys::Event60(Client *client) {

  this->P(61,client,1);
}

// CLASS IOSubSys: Event #1 - Reservation of processor

void IOSubSys::Event61(Client *client) {

  rsproc->P(62,client,1);
}

// CLASS IOSubSys: Event #2 - Reservation of disk

void IOSubSys::Event62(Client *client) {

  client->pageindex=1;          // 1st page to access
  rsdisk->P(63,client,1);
}

// CLASS IOSubSys: Event #3 - Disk access

void IOSubSys::Event63(Client *client) {

  if (client->pageindex<=client->pgnums->NbElem()) {
    AccDisk(client);                    // R5
    // Stats
    NBIO++;
    if (client->type<=4) TNBIO++;
  } else { // END
    rsdisk->V();                        // Release disk
    rsproc->V();                        // Release processor
    this->V();                          // Release AR5
    if (client->type!=7) // Read operation => page in memory
      Sim()->Sched()->Schedule(53,Sim()->Tnow(),client);
    else // Write operation d'ecriture => END of treatment, client is destroyed
      Sim()->KillClient(client);
  }
}

// CLASS IOSubSys: Disk access (R5)

void IOSubSys::AccDisk(Client *client) {

  int pagetoaccess;

  pagetoaccess=client->pgnums->Scan(client->pageindex++);
//  printf("Disk access (");
//  if (client->type!=7) printf("read");
//  else printf("write");
//  printf(") to page #%d\n",pagetoaccess);
  if ((pagetoaccess<(lastpage-1)) || (pagetoaccess>(lastpage+1))) {
    lastpage=pagetoaccess;
    // Non-contiguous access
    Sim()->Sched()->Schedule(63,Sim()->Tnow()+DISKSEA+DISKLAT+DISKTRA,client);
  } else {
    lastpage=pagetoaccess;
    // Contiguous access: no search time
    Sim()->Sched()->Schedule(63,Sim()->Tnow()+DISKLAT+DISKTRA,client);
  }
}

