
/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database 2)
// Generic performance evaluation model of OODBs
// Version 2.0 C++, April 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// voodb2.cc: Main program
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Includes
/////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include "simutil.h"
#include "voodb2.h"
#include "voodb2uc.h"
#include "basec.h"
Database *DB;                 // Database object (global variable)
#include "simulc.h"
#include "eventc.h"
#include "voodb2um.h"
#include "basem.h"
#include "simulm.h"
#include "eventm.h"

/////////////////////////////////////////////////////////////////////
// Prototypes
/////////////////////////////////////////////////////////////////////

void GetParam();                // Parameters interactive input
void DisplayParam();            // Parameters display
void ParamTest();               // Parameters testing

/////////////////////////////////////////////////////////////////////
// Main program
/////////////////////////////////////////////////////////////////////

void main(void) {

  char c;
  int i,j;
  Simulation *sim;

  // Intro.
  printf("\n*** VOODB2 (C++) ***\n");

  // Parameters input
  printf("\n(D)efault parameters / (C)ustom parameters : ");
  c=getchar();
  if ((c!='D') && (c!='d')) GetParam();
  DisplayParam();
  ParamTest();

  // Induced parameters computation

  PCT1=PSET;
  PCT2=PCT1+PSIMPLE;
  PCT3=PCT2+PHIER;
  PCC=PCT3+PSTOCH;

  j=2;
  STOCHTAB[0]=1.0/j;
  for (i=1; i<MAXNREF; i++) {
    j=j*2;
    STOCHTAB[i]=STOCHTAB[i-1]+1.0/j;
  }

  // Instantiation of object simulation

  sim=new Simulation(0,TSIM,-1);

  // Simulation execution

  sim->Run(NREPLIC);

  // Clean-up

  delete sim;
}

/////////////////////////////////////////////////////////////////////
// Interactive parameters input
/////////////////////////////////////////////////////////////////////

void GetParam() {

  printf("\nTSIM=");scanf("%f",&TSIM);
  printf("NREPLIC=");scanf("%d",&NREPLIC);

  printf("\nSYSCLASS=");scanf("%d",&SYSCLASS);
  printf("MULTILVL=");scanf("%d",&MULTILVL);
  printf("NUSERS=");scanf("%d",&NUSERS);
  printf("PGSIZE=");scanf("%d",&PGSIZE);
  printf("BUFFSIZE=");scanf("%d",&BUFFSIZE);
  printf("GETLOCK=");scanf("%f",&GETLOCK);
  printf("RELLOCK=");scanf("%f",&RELLOCK);
  printf("DISKSEA=");scanf("%f",&DISKSEA);
  printf("DISKLAT=");scanf("%f",&DISKLAT);
  printf("DISKTRA=");scanf("%f",&DISKTRA);
  printf("NETTHRU=");scanf("%f",&NETTHRU);

  printf("\nCLUSTP=");scanf("%d",&CLUSTP);
  printf("INITP=");scanf("%d",&INITP);
  printf("PGREP=");scanf("%d",&PGREP);
  printf("KLRU=");scanf("%d",&KLRU);
  printf("PREFETCH=");scanf("%d",&PREFETCH);

  printf("\nNC=");scanf("%d",&NC);
  printf("MAXNREF=");scanf("%d",&MAXNREF);
  printf("BASESIZE=");scanf("%d",&BASESIZE);
  printf("NO=");scanf("%d",&NO);
  printf("NREFT=");scanf("%d",&NREFT);
  printf("DIST1=");scanf("%d",&DIST1);
  printf("DIST2=");scanf("%d",&DIST2);
  printf("DIST3=");scanf("%d",&DIST3);
  printf("DIST4=");scanf("%d",&DIST4);
  printf("INFCLASS=");scanf("%d",&INFCLASS);
  printf("SUPCLASS=");scanf("%d",&SUPCLASS);

  printf("\nTHINK=");scanf("%f",&THINK);
  printf("PSET=");scanf("%f",&PSET);
  printf("PSIMPLE=");scanf("%f",&PSIMPLE);
  printf("PHIER=");scanf("%f",&PHIER);
  printf("PSTOCH=");scanf("%f",&PSTOCH);
  printf("PCLUSTD=");scanf("%f",&PCLUSTD);
  printf("SETDEPTH=");scanf("%d",&SETDEPTH);
  printf("SIMDEPTH=");scanf("%d",&SIMDEPTH);
  printf("HIEDEPTH=");scanf("%d",&HIEDEPTH);
  printf("STODEPTH=");scanf("%d",&STODEPTH);
  printf("DIST5=");scanf("%d",&DIST5);

  printf("\nDSTC\nN=");scanf("%d",&N);
  printf("NP=");scanf("%d",&NP);
  printf("P=");scanf("%d",&P);
  printf("TFA=");scanf("%f",&TFA);
  printf("TFE=");scanf("%f",&TFE);
  printf("TFC=");scanf("%f",&TFC);
  printf("W=");scanf("%f",&W);
  printf("S=");scanf("%d",&S);
  printf("\n");
}

/////////////////////////////////////////////////////////////////////
// Parameters display
/////////////////////////////////////////////////////////////////////

void DisplayParam() {

  printf("\n*** Simulation parameters ***\n\n");

  printf("     TSIM=%f\n",TSIM);
  printf("  NREPLIC=%d\n\n",NREPLIC);

  printf(" SYSCLASS=%d\n",SYSCLASS);
  printf(" MULTILVL=%d\n",MULTILVL);
  printf("   NUSERS=%d\n",NUSERS);
  printf("   PGSIZE=%d\n",PGSIZE);
  printf(" BUFFSIZE=%d\n",BUFFSIZE);
  printf("  GETLOCK=%f\n",GETLOCK);
  printf("  RELLOCK=%f\n",RELLOCK);
  printf("  DISKSEA=%f\n",DISKSEA);
  printf("  DISKLAT=%f\n",DISKLAT);
  printf("  DISKTRA=%f\n",DISKTRA);
  printf("  NETTHRU=%f\n\n",NETTHRU);

  printf("   CLUSTP=%d\n",CLUSTP);
  printf("    INITP=%d\n",INITP);
  printf("    PGREP=%d\n",PGREP);
  printf("     KLRU=%d\n",KLRU);
  printf(" PREFETCH=%d\n\n",PREFETCH);

  printf("       NC=%d\n",NC);
  printf("  MAXNREF=%d\n",MAXNREF);
  printf(" BASESIZE=%d\n",BASESIZE);
  printf("       NO=%d\n",NO);
  printf("    NREFT=%d\n",NREFT);
  printf("    DIST1=%d\n",DIST1);
  printf("    DIST2=%d\n",DIST2);
  printf("    DIST3=%d\n",DIST3);
  printf("    DIST4=%d\n",DIST4);
  printf(" INFCLASS=%d\n",INFCLASS);
  printf(" SUPCLASS=%d\n\n",SUPCLASS);

  printf("    THINK=%f\n",THINK);
  printf("     PSET=%f\n",PSET);
  printf("  PSIMPLE=%f\n",PSIMPLE);
  printf("    PHIER=%f\n",PHIER);
  printf("   PSTOCH=%f\n",PSTOCH);
  printf("  PCLUSTD=%f\n",PCLUSTD);
  printf(" SETDEPTH=%d\n",SETDEPTH);
  printf(" SIMDEPTH=%d\n",SIMDEPTH);
  printf(" HIEDEPTH=%d\n",HIEDEPTH);
  printf(" STODEPTH=%d\n",STODEPTH);
  printf("    DIST5=%d\n\n",DIST5);

  printf("DSTC    N=%d\n",N);
  printf("       NP=%d\n",NP);
  printf("        P=%d\n",P);
  printf("      TFA=%f\n",TFA);
  printf("      TFE=%f\n",TFE);
  printf("      TFC=%f\n",TFC);
  printf("        W=%f\n",W);
  printf("        S=%d\n",S);
}

/////////////////////////////////////////////////////////////////////
// Parameters testing
/////////////////////////////////////////////////////////////////////

void ParamTest() {

  short halt=0;

  if (MAXNREF>MMAXNREF) {
    printf("\nMAXNREF value (%d) is too large! It should be <= %d.\n",
      MAXNREF,MMAXNREF);
    halt=1;
  }
  if (NUSERS>MNUSERS) {
    printf("\nNUSERS value (%d) is too large! It should be <= %d.\n",
      NUSERS,MNUSERS);
    halt=1;
  }
  if (NUSERS>INFINITE) {
    printf("\nNUSERS value (%d) is too large! It should be <= %d (INFINITE).\n",
      NUSERS,INFINITE);
    halt=1;
  }
  if (NC>MNC) {
    printf("\nNC value (%d) is too large! It should be <= %d.\n",
      NC,MNC);
    halt=1;
  }
  if (NO>MNO) {
    printf("\nNO value (%d) is too large! It should be <= %d.\n",
      NO,MNO);
    halt=1;
  }
  if (NREFT>MNREFT) {
    printf("\nNREFT value (%d) is too large! It should be <= %d.\n",
      NREFT,MNREFT);
    halt=1;
  }
  if (KLRU>MKLRU) {
    printf("\nKLRU value (%d) is too large! It should be <= %d.\n",
      KLRU,MKLRU);
    halt=1;
  }
  if (BUFFSIZE>MBUFFSIZE) {
    printf("\nBUFFSIZE value (%d) is too large! It should be <= %d.\n",
      BUFFSIZE,MBUFFSIZE);
    halt=1;
  }

  if (halt) {
    puts("\n*** STOP ***\n");
    exit(-1);
  }
}

