
/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database 2)
// Generic performance evaluation model of OODBs
// Version 2.0 C++, April 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// voodb2.h:   - Globale variables definition
//             - Utilitary functions definition
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Constants
/////////////////////////////////////////////////////////////////////

#define MMAXNREF  10            // Superior bound for MAXNREF
#define MNUSERS   10            // Superior bound for NUSERS
#define INFINITE  99            // Must be greater than MNUSERS
#define MNC       50            // Superior bound for NC
#define MNO       20000         // Superior bound for NO
#define MNREFT    5             // Superior bound for NREFT
#define MKLRU     5             // Superior bound for KLRU
#define MBUFFSIZE 10000         // Superior bound for BUFFSIZE
#define NTTRANS   4             // Number of transaction types

/////////////////////////////////////////////////////////////////////
// Simulation parameters with default values
/////////////////////////////////////////////////////////////////////

// Simulation control

float TSIM=4000000;             // Simulation time
int NREPLIC=10;                 // Number of replications

// System definition

short SYSCLASS=3;               // System class
                                // 1 = Centralized
                                // 2 = Object server
                                // 3 = Page server
                                // 4 = Database server
                                // 9 = Other
int MULTILVL=10;                // Multiprogramming level
int NUSERS=1;                   // Number of users
int PGSIZE=4096;                // Disk page size
int BUFFSIZE=100;               // Buffer size (in pages)
float GETLOCK=0.5;              // Locks acquisition time
float RELLOCK=0.5;              // Locks release time
float DISKSEA=18;               // Disk search time (in ms)
float DISKLAT=6;                // Disk latency time (in ms)
float DISKTRA=1;                // Disk transfer time (in ms)
float NETTHRU=1024;             // Network throughput (in KB/s)

// Definition of components influencing the performance

short CLUSTP=9;                 // Clustering policy
                                // 0 = None
                                // 9 = Other (DSTC)
short INITP=2;                  // Objects initial placement
                                // 1 = Sequential
                                // 2 = Optimized sequential
                                // 9 = Other
short PGREP=4;                  // Buffer page replacement policy
                                // 1 = RANDOM
                                // 2 = FIFO
                                // 3 = LFU
short KLRU=1;                   // 4 = LRU-K (KLRU, parameter K of LRU)
                                // 5 = CLOCK
                                // 6 = GCLOCK
                                // 9 = Other
short PREFETCH=0;               // Object prefetching policy
                                // 0 = None
                                // 9 = Other (DSTC)

// OCB (Object Clustering Benchmark) database definition

int NC=50;                      // Number of classes
short MAXNREF=10;               // Max. number of references
int BASESIZE=50;                // Instance base size
int NO=20000;                   // Total number of objects
short NREFT=4;                  // Number of reference types
short DIST1=1;                  // Reference types random distribution
short DIST2=1;                  // Class references random distribution
short DIST3=1;                  // Objects within classes random distribution
short DIST4=1;                  // Object references random distribution
                                // 1 = Uniform
                                // 2 = Normal
                                // 9 = Other
int INFCLASS=1;                 // Inferior bound for classes referenced by DIST2
int SUPCLASS=NC;                // Superior bound for classes referenced by DIST2

// OCB transactions definition 

float THINK=4000;               // Mean latency time between two transactions
float PSET=0.25;                // Set-oriented access occurrence probability
float PSIMPLE=0.25;             // Simple traversal occurrence probability
float PHIER=0.25;               // Hierarchy traversal occurrence probability
float PSTOCH=0.25;              // Stochastic traversal occurrence probability
float PCLUSTD=0;                // External clustering demand occurrence probability
int SETDEPTH=3;                 // Set-oriented accesses depth
int SIMDEPTH=3;                 // Simple traversals depth
int HIEDEPTH=5;                 // Hierarchy traversals depth
int STODEPTH=50;                // Stochastic traversals depth
short DIST5=1;                  // Transaction root objects random distribution

// DSTC parameters

int N=200;                      // Maximum number of entries in MO during one period
int NP=200;                     // Number of periods after which a link is obsolete
int P=1000;                     // Period modulo
float TFA=1.0;                  // Tfa threshold value
float TFE=1.0;                  // Tfe threshold value
float TFC=1.0;                  // Tfc threshold value
float W=0.3;                    // Weight
int S=10;                       // Prefetching window size

// Induced parameters 

float PCT1, PCT2, PCT3, PCC;    // Cumulated occurrence probabilities for
                                // transactions and clustering demands
float STOCHTAB[MMAXNREF];       // Stochastic traversal probability table

/////////////////////////////////////////////////////////////////////
// VOODB2-specific statistical counters
/////////////////////////////////////////////////////////////////////

// Number of transactions

int NBTRANS[NTTRANS+1];
float NBTRANSC[NTTRANS+1], NBTRANSC2[NTTRANS+1];

// Response time

float RTIME[NTTRANS+1], RTIMEC[NTTRANS+1], RTIMEC2[NTTRANS+1];
int NBST[NTTRANS+1]; // # Started transactions

// Number of accessed objects

int NBOBJ[NTTRANS+1];
float NBOBJC[NTTRANS+1], NBOBJC2[NTTRANS+1];

// Number of reclustering operations

int NBCL;
float NBCLC, NBCLC2;

// Clustering duration

float CTIME, CTIMEC, CTIMEC2;

// Clustering overhead

float TCTIME, TCTIMEC, TCTIMEC2;

// Number of I/Os (total, transactions, clustering)

int NBIO, TNBIO, CNBIO;
float NBIOC, NBIOC2, TNBIOC, TNBIOC2, CNBIOC, CNBIOC2;

// Number of disk pages

float NBPGC, NBPGC2;

// Number and average size of clusters

int NCLUSTERS, TCLUSTERS;
float NCC, NCC2, TCC, TCC2;

/////////////////////////////////////////////////////////////////////
// Customized random integer generation
/////////////////////////////////////////////////////////////////////
// Returns an integer belongin to the [min, max] interval depending
// on the dist random distribution
/////////////////////////////////////////////////////////////////////

int CRand(short dist, int min, int max) {

  int res;

  switch(dist) {

    // Uniform distribution 
    case 1: res=IUni(min,max);break;

    // Normal distribution 
    case 2: res=-1;break; // TO DO!!!

    // Other distribution (user-defined)
    case 9: res=-1;break;

    // Erreur
    default: res=-1;printf("\nError: Unknown random distribution!\n");

  }

  return res;
}
