
/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database 2)
// Generic performance evaluation model of OODBs
// Version 2.0 C++, April 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// voodb2um.h: Utilitary methods definition
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// CLASS ChList
/////////////////////////////////////////////////////////////////////

// CLASS ChList: Constructor

ChList::ChList() {

  head=NULL;
  queue=NULL;
  nbelem=0;
  position=0;
  cur=NULL;
  prev=NULL;
  next=NULL;
}

// CLASS ChList: Destructor

ChList::~ChList() {

  IntCell *del;

  cur=head;
  while (cur!=NULL) {
    del=cur;
    cur=cur->Next();
    delete del;
  }
}

// CLASS ChList: Pointers display

void ChList::Print() {

  printf("\n\t     nbelem = %d\n",nbelem);
  printf("\t head value = ");
  if (head!=NULL) printf("%d\n",head->Value());
  else printf("NULL\n");
  printf("\tqueue value = ");
  if (queue!=NULL) printf("%d\n",queue->Value());
  else printf("NULL\n");
  printf("\t   position = %d\n",position);
  printf("\t  cur value = ");
  if (cur!=NULL) printf("%d\n",cur->Value());
  else printf("NULL\n");
  printf("\t prev value = ");
  if (prev!=NULL) printf("%d\n",prev->Value());
  else printf("NULL\n");
  printf("\t next value = ");
  if (next!=NULL) printf("%d\n",next->Value());
  else printf("NULL\n");
}

// CLASS ChList: Positioning 

// Modified attributes in output: position, cur, prev, next

void ChList::Pos(int place) {

  int dist1, dist2, dist3;

  // Distance to resp. head, pos, and end

  dist1=abs(place-1);
  dist2=abs(position-place);
  dist3=abs(nbelem-place);

  // Positioning in the list

  if ((dist1<dist2)&&(dist1<dist3)) { // Head
    position=1;
    cur=head;
  } else if ((dist3<dist2)&&(dist3<dist1)) { // End
    position=nbelem;
    cur=queue;
  } // Else stay here

  // Selection of forward/backward way

  if ((cur==head)||(place>=position)) { // "Forward"
    if (cur!=NULL) prev=cur->Prev();
    else prev=NULL;
    while ((cur!=NULL) && (position<place)) {
      prev=cur;
      cur=cur->Next();
      position++;
    }
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
  } else { // "Backward"
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
    while ((cur!=NULL) && (position>place)) {
      next=cur;
      cur=cur->Prev();
      position--;
    }
    if (cur!=NULL) prev=cur->Prev();
    else prev=NULL;
  }
}

// CLASS ChList: Insertion 

void ChList::Insert(int place, int value) {

  IntCell *newcell;

  // Positioning

  Pos(place);

  // Insert new cell

  newcell=new IntCell(value);
  newcell->SetNext(cur);
  newcell->SetPrev(prev);
  if (cur!=NULL) {
    cur->SetPrev(newcell);
    if (prev!=NULL) prev->SetNext(newcell);
    else head=newcell;          // Insertion - head
  } else {
    if (position==1) {          // 1st element
      head=newcell;
      queue=newcell;

    } else {                    // Insertion - end
      prev->SetNext(newcell);
      queue=newcell;
    }
    next=NULL;
  }
  nbelem++;
  cur=newcell;
  prev=cur->Prev();
  next=cur->Next();
}

// CLASS ChList: Update

void ChList::Update(int place, int value) {

  // Positioning
  
  Pos(place);

  // Update value of current cell
  
  if (cur!=NULL) cur->SetVal(value);
  else { // Repositioning on the head
    position=1;
    cur=head;
    prev=NULL;
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
  }
}

// CLASS ChList: Deletion

void ChList::Delete(int place) {

  // Positioning

  Pos(place);

  // Delete current cell

  if (cur!=NULL) {
    if (prev!=NULL) prev->SetNext(next);
    else  head=next;             // Delete head
    if (next!=NULL) next->SetPrev(prev);
    else {             // Delete end
      queue=prev;
      position--;
    }
    delete cur;
    nbelem--;
    if (position>1) cur=prev;
    else cur=head;
    if (cur!=NULL) {
      prev=cur->Prev();
      next=cur->Next();
    } else {
      next=NULL;
      prev=NULL;
      position=0;
    }
  } else { // Repositioning on the head
    position=1;
    cur=head;
    prev=NULL;
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
  }
}

// CLASS ChList: Chained list reversal

void ChList::Reverse() {

  IntCell *newhead, *newqueue, *lcur, *newcell;

  // Copy

  newhead=NULL;
  newqueue=NULL;
  lcur=head;
  while (lcur!=NULL) {
    // New cell inserted at the beginning (head) of the new list
    newcell=new IntCell(lcur->Value());
    newcell->SetNext(newhead);
    newcell->SetPrev(NULL);
    newhead->SetPrev(newcell);
    if (newhead==NULL) newqueue=newcell;
    newhead=newcell;
    lcur=lcur->Next();
  }

  // Deletion of old list & pointer assignation

  lcur=head;
  while (lcur!=NULL) {
    newcell=lcur;
    lcur=lcur->Next();
    delete newcell;
  }
  head=newhead;
  queue=newqueue;
  position=1;
  cur=head;
  prev=NULL;
  next=cur->Next();
}

// CLASS ChList: Return value

int ChList::Scan(int place) {

  // Positioning
  Pos(place);

  // Result
  if ((cur!=NULL) && (position==place)) return cur->Value();
  else { // Repositioning on head
    position=1;
    cur=head;
    prev=NULL;
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
    return -1;
  }
}

// CLASS ChList: Return the number of elements

int ChList::NbElem() {

  return nbelem;
}

// CLASS ChList: Display the whole list

void ChList::Display() {

  IntCell *lcur;

  printf("LIST (%d elements):",nbelem);
  lcur=head;
  while (lcur!=NULL) {
    printf(" %d",lcur->Value());
    lcur=lcur->Next();
  }
  printf("\n");
}

/////////////////////////////////////////////////////////////////////
// CLASS IntCell
/////////////////////////////////////////////////////////////////////

// CLASS IntCell: Constructor

IntCell::IntCell(int val) {

  value=val;
}

// CLASS IntCell: Value initialization

void IntCell::SetVal(int val) {

  value=val;
}

// CLASS IntCell: Next pointer initialization

void IntCell::SetNext(IntCell *n) {

  next=n;
}

// CLASS IntCell: Prev pointer initialization

void IntCell::SetPrev(IntCell *n) {

  prev=n;
}

// CLASS IntCell: Return value

int IntCell::Value() {

  return value;
}

// CLASS IntCell: Return next pointer

IntCell* IntCell::Next() {

  return next;
}

// CLASS IntCell: Return prev pointer

IntCell* IntCell::Prev() {

  return prev;
}
